# -*- coding: utf-8 -*-

# Aqua-Duct, a tool facilitating analysis of the flow of solvent molecules in molecular dynamic simulations
# Copyright (C) 2018-2019  Tomasz Magdziarz, Michał Banas <info@aquaduct.pl>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from aquaduct import logger
# import logging
# logger = logging.getLogger(__name__)

from aquaduct.utils.clui import SimpleTree
import pickle as pickle
import gzip
import os
from collections import OrderedDict

from aquaduct.apps.data import GCS

try:
    if GCS.nc4:
        import netCDF4 as netcdf

        logger.debug('NetCDF AQ format: Imported netCDF4')
    else:
        raise ImportError()
except ImportError:
    from scipy.io import netcdf

    logger.debug('NetCDF AQ format: Imported scipy.io.netcdf')

import numpy as np

from aquaduct import version
from aquaduct.utils.helpers import dictify

from aquaduct.traj.sandwich import ResidueSelection
from aquaduct.traj.paths import GenericPaths, SinglePath, PassingPath, SinglePathID, GenericPathTypeCodes, MasterPath
from aquaduct.utils.helpers import SmartRange
from aquaduct.traj.inlets import InletClusterExtendedType, InletClusterGenericType, Inlet, Inlets
from aquaduct.traj.barber import Sphere
from aquaduct.geom.master import FakeSingleResidueSelection

from itertools import chain
import json


################################################################################
# Version checking

def check_version_compliance(current, loaded, what):
    if current[0] > loaded[0]:
        logger.error('Loaded data has %s major version lower then the application.' % what)
    if current[0] < loaded[0]:
        logger.error('Loaded data has %s major version higher then the application.' % what)
    if current[0] != loaded[0]:
        logger.error('Possible problems with API compliance.')
    if current[1] > loaded[1]:
        logger.warning('Loaded data has %s minor version lower then the application.' % what)
    if current[1] < loaded[1]:
        logger.warning('Loaded data has %s minor version higher then the application.' % what)
    if current[1] != loaded[1]:
        logger.warning('Possible problems with API compliance.')


def check_versions(version_dict):
    assert isinstance(version_dict, (dict, OrderedDict)), "File is corrupted, cannot read version data."
    assert 'version' in version_dict, "File is corrupted, cannot read version data."
    assert 'aquaduct_version' in version_dict, "File is corrupted, cannot read version data."
    check_version_compliance(version(), version_dict['aquaduct_version'], 'Aqua-Duct')
    check_version_compliance(version(), version_dict['version'], 'Valve')


################################################################################

def get_vda_reader(filename, mode="r"):
    if os.path.splitext(filename)[-1].lower() in ['.dump']:
        return ValveDataAccess_pickle(mode=mode, data_file_name=filename)
    elif os.path.splitext(filename)[-1].lower() in ['.npaq']:
        return ValveDataAccess_numpy(mode=mode, data_file_name=filename)
    elif os.path.splitext(filename)[-1].lower() in ['.nc', '.aqnc']:
        return ValveDataAccess_nc(mode=mode, data_file_name=filename)
    raise ValueError('Unknown file type of %s file' % filename)


################################################################################


class ValveDataAccess(object):
    unknown_names = 'UNK'

    def __init__(self, mode=None, data_file_name=None):
        logger.debug('Opening file %s in mode %s' % (data_file_name, mode))
        self.data_file_name = data_file_name
        self.data_file = None
        self.data = None
        assert mode in 'rw'
        self.mode = mode  # r, w
        self.open()

    def open(self):
        raise NotImplementedError()

    def close(self):
        raise NotImplementedError()

    def __del__(self):
        self.close()

    def load(self):
        raise NotImplementedError()

    def dump(self, **kwargs):
        raise NotImplementedError()

    def get_variable(self, name):
        raise NotImplementedError()

    def set_variable(self, name, value):
        raise NotImplementedError()


################################################################################


class ValveDataCodec(object):
    '''
    Class defines format of encoding AQ objects into NetCDF format.
    Encoding defined here can be also used to store AQ objects as NumPy arrays.
    '''

    # this is in fact definition of data format
    # it assumes data is dictionary with scipy netcdf variables, would it work with netcdf4?

    version = 0, 0, 1
    '''
    Current version of :class:`ValveDataCodec`
    '''

    @staticmethod
    def varname(name, *suffix):
        '''
        Name of variable made by combining base name and suffixes (if any).
        Base name and suffixes are joined with dot '.'.

        :param name: Base for variable name.
        :param suffix: Optional suffixes.
        :return: Name of variable made by combining base name and suffixes (if any).
        Base name and suffixes are joined with dot '.'.
        '''
        if len(suffix):
            suff = '.'.join(map(str, suffix))
            return '%s.%s' % (name, suff)
        return '%s' % name

    @staticmethod
    def encode(name, value):

        if name == 'center_of_system':
            # center_of_system: (3,)*float

            yield name, np.array(value)

        if name == 'all_res':
            # all_res.layers: (L,)*int
            #   L is a number of layers
            # all_res.layer.N: (S,)*int
            #   N is a consecutive layer number
            #   S is a number of traced molecules in N layer

            # save layers
            layers = np.array(list(value.selected.keys()), dtype=np.int32)
            yield ValveDataCodec.varname(name, 'layers'), layers
            # save each layer as separate array
            for l in layers:
                yield ValveDataCodec.varname(name, 'layer', l), np.array(value.selected[l], dtype=np.int32)

        if name == 'number_frame_rid_in_object':
            # number_frame_rid_in_object.layers: (1,)*int
            #   number of layers
            # number_frame_rid_in_object.layer.N.sizes: (F,)*int
            #   N is a consecutive layer number
            #   F is a number of frames in layer N
            # number_frame_rid_in_object.layer.N: (Q,)*int
            #   Q is a number of molecules in in object in all frames,
            #   it is a sum of number_frame_rid_in_object.layer.N.sizes

            # TODO It can be further improved because there is no sense to store repeated information for frames.
            # number of layers
            layers = len(value)
            yield ValveDataCodec.varname(name, 'layers'), np.array([layers], dtype=np.int32)
            for nr, layer in enumerate(value):
                # for each layer make sizes array
                yield ValveDataCodec.varname(name, 'layer', nr, 'sizes'), np.fromiter((len(row) for row in layer),
                                                                                      dtype=np.int32)
                # then save layer
                yield ValveDataCodec.varname(name, 'layer', nr), np.fromiter(chain(*(row for row in layer)),
                                                                             dtype=np.int32)

        if name == 'paths':
            # paths.layers:  (L,)*int
            #   L is a number of layers
            # paths.layer.N.names: (P,3)*str
            #   N is a consecutive layer number
            #   P is a number of paths in N layer
            # paths.layer.N.ids: (P,)*int
            #   N is a consecutive layer number
            #   P is a number of paths in N layer
            # paths.layer.N.min_max_frames: (1,2)*int
            # paths.layer.N.object.sizes: (P,)*int
            #   P is a number of paths in N layer
            # paths.layer.N.scope.sizes: (P,)*int
            #   P is a number of paths in N layer
            # paths.layer.N.object: (PO*2,)*int
            #   PO is a number of pairs (start,times) decoding object paths for all paths in N layer
            # paths.layer.N.scope: (PS*2,)*int
            #   PO is a number of pairs (start,times) decoding scope paths for all paths in N layer

            # get all layers
            layers = sorted(set((p.id[0] for p in value)))
            yield ValveDataCodec.varname(name, 'layers'), np.array(layers, dtype=np.int32)
            for N in layers:
                P = sum((1 for p in value if p.id[0] == N))
                yield ValveDataCodec.varname(name, 'layer', N, 'names'), np.fromiter(
                    chain(*(p.name for p in value if p.id[0] == N)), dtype='S1').reshape(P, 3)
                yield ValveDataCodec.varname(name, 'layer', N, 'ids'), np.fromiter(
                    (p.id[-1] for p in value if p.id[0] == N), dtype=np.int32)
                mmf = next((p for p in value if p.id[0] == N))
                yield ValveDataCodec.varname(name, 'layer', N, 'min_max_frames'), np.array(
                    (mmf.min_possible_frame, mmf.max_possible_frame), dtype=np.int32)
                oors = [SmartRange(p.frames_of_object) for p in value if p.id[0] == N]
                yield ValveDataCodec.varname(name, 'layer', N, 'object', 'sizes'), np.fromiter(
                    (2 * len(list(o.raw_increment)) for o in oors), dtype=np.int32)
                yield ValveDataCodec.varname(name, 'layer', N, 'object'), np.fromiter(
                    chain(*(o.raw2sequence(o.raw_increment) for o in oors)), dtype=np.int32)
                oors = [SmartRange(p.frames_of_scope) for p in value if p.id[0] == N]
                yield ValveDataCodec.varname(name, 'layer', N, 'scope', 'sizes'), np.fromiter(
                    (2 * len(list(o.raw_increment)) for o in oors), dtype=np.int32)
                yield ValveDataCodec.varname(name, 'layer', N, 'scope'), np.fromiter(
                    chain(*(o.raw2sequence(o.raw_increment) for o in oors)), dtype=np.int32)

        if name == 'spaths':
            # spaths.layers:  (L,)*int
            #   L is a number of layers
            # spaths.layer.N.names: (P,3)*str
            #   N is a consecutive layer number
            #   P is a number of paths in N layer
            # spaths.layer.N.ids: (P,2)*int
            #   N is a consecutive layer number
            #   P is a number of paths in N layer
            # spaths.layer.N.single: (P,)*int
            #   P is a number of paths in N layer
            # spaths.layer.N.frames: (P,5)*int
            #   P is a number of paths in N layer
            #   table of (begining,end,in,object,out)
            # spaths.layer.N.object.sizes: (PSO,)*int
            #   PSO is a number of paths in N layer that are not PassingPaths
            # spaths.layer.N.object: (PSOS*2,)*int
            #   PSOS is a number of pairs (start,times) decoding strict obejct frames of non PassingPaths for all paths in N layer

            layers = sorted(set((p.id.id[0] for p in value)))
            yield ValveDataCodec.varname(name, 'layers'), np.array(layers, dtype=np.int32)
            for N in layers:
                P = sum((1 for p in value if p.id.id[0] == N))  # number of paths in N
                yield ValveDataCodec.varname(name, 'layer', N, 'names'), np.fromiter(
                    chain(*(p.id.name for p in value if p.id.id[0] == N)), dtype='S1').reshape(P, 3)
                yield ValveDataCodec.varname(name, 'layer', N, 'ids'), np.fromiter(
                    chain(*((p.id.id[-1], p.id.nr) for p in value if p.id.id[0] == N)), dtype=np.int32).reshape(P, 2)
                yield ValveDataCodec.varname(name, 'layer', N, 'single'), np.fromiter(
                    (p.is_single() for p in value if p.id.id[0] == N), dtype='i1')  # i1 instead of bool
                yield ValveDataCodec.varname(name, 'layer', N, 'frames'), np.fromiter(
                    chain(*((p.begins, p.ends) + tuple(p.sizes) for p in value if p.id.id[0] == N)),
                    dtype=np.int32).reshape(P, 5)
                osf = [SmartRange(p.path_object_strict()) for p in value if p.id.id[0] == N and p.is_single()]
                yield ValveDataCodec.varname(name, 'layer', N, 'object', 'sizes'), np.fromiter(
                    (2 * len(list(o.raw_increment)) for o in osf), dtype=np.int32)
                yield ValveDataCodec.varname(name, 'layer', N, 'object'), np.fromiter(
                    chain(*(o.raw2sequence(o.raw_increment) for o in osf)), dtype=np.int32)

        if name == 'ctypes':
            # ctypes: (P,4)*int
            #   None is replaced by -1

            yield ValveDataCodec.varname(name), np.fromiter(
                chain(*([-1 if c is None else c for c in ct.clusters] for ct in value)), dtype=np.int32).reshape(
                len(value), 4)

        if name in ['master_paths', 'master_paths_smooth']:
            # master_paths*.keys: (M,2)*int
            #   M is a number of master paths
            #   keys are ctypes generic
            # master_paths*.names: (M,3)*str
            #   M is a number of master paths
            # master_paths*.ids: (M,3)*int
            #   M is a number of master paths
            # master_paths*.frames: (M,5)*int
            #   M is a number of master paths
            #   table of (begining,end,in,object,out)
            # master_paths*.widths: (MS,)*float
            #   MS is a sum of lenghts of all master paths
            # master_paths*.coords: (MS,3)*int
            #   MS is a sum of lenghts of all master paths
            # master_paths*.object.sizes: (M,)*int
            #   M is a number of master paths
            # master_paths*.object: (MOS*2,)*int
            #   MOS is a number of pairs (start,times) decoding strict obejct frames of master paths

            M = len(value)
            yield ValveDataCodec.varname(name, 'keys'), np.fromiter(
                chain(*([-1 if c is None else c for c in ct.clusters] for ct in list(value.keys()))),
                dtype=np.int32).reshape(len(value), 2)
            yield ValveDataCodec.varname(name, 'names'), np.fromiter(chain(*(p.id.name for p in list(value.values()))),
                                                                     dtype='S1').reshape(M, 3)
            yield ValveDataCodec.varname(name, 'ids'), np.fromiter(
                chain(*((p.id.id[0], p.id.id[-1], p.id.nr) for p in list(value.values()))), dtype=np.int32).reshape(M, 3)
            yield ValveDataCodec.varname(name, 'frames'), np.fromiter(
                chain(*((p.begins, p.ends) + tuple(p.sizes) for p in list(value.values()))), dtype=np.int32).reshape(M, 5)
            yield ValveDataCodec.varname(name, 'widths'), np.fromiter(chain(*(p.width_cont for p in list(value.values()))),
                                                                      dtype=np.float32)
            MS = sum((p.size for p in list(value.values())))
            yield ValveDataCodec.varname(name, 'coords'), np.fromiter(
                chain(*chain(*(p.coords_cont for p in list(value.values())))), dtype=np.float32).reshape(MS, 3)
            osf = [SmartRange(p.path_object_strict()) for p in list(value.values())]
            yield ValveDataCodec.varname(name, 'object', 'sizes'), np.fromiter(
                (2 * len(list(o.raw_increment)) for o in osf), dtype=np.int32)
            yield ValveDataCodec.varname(name, 'object'), np.fromiter(
                chain(*(o.raw2sequence(o.raw_increment) for o in osf)), dtype=np.int32)

        if name == 'inls':
            # inls is a complicated object
            #
            # inls.center_of_system: (3,)*float
            # inls.onlytype: (1,)*str
            # inls.inlets_list.coords: (I,3)*float
            # inls.inlets_list.frame: (I,)*int
            # inls.inlets_list.type: (I,)*int
            #   it references onlytype
            # inls.inlets_list.reference.ids: (I,3)*int
            # inls.inlets_list.reference.name: (I,3)*str
            # inls.inlets_ids: (I,)*int
            #   I is a number of inlets
            # inls.clusters: (I,)*int
            #   I is a number of inlets
            # inls.number_of_clustered_inlets: (1,)*int
            # inls.spheres: (I,4)*float
            #   I is a number of inlets
            # inls.spheres.nr: (I,)*int
            #   I is a number of inlets
            # inls.passing: (1,)*int
            # inls.tree: (1,)*str

            if value.center_of_system is not None:
                yield ValveDataCodec.varname(name, 'center_of_system'), np.array(value.center_of_system)
            # yield ValveDataCodec.varname(name, 'onlytype'), np.fromiter(json.dumps(value.onlytype),dtype='S1')
            yield ValveDataCodec.varname(name, 'inlets_list', 'coords'), np.array([i.coords for i in value.inlets_list])
            yield ValveDataCodec.varname(name, 'inlets_list', 'frame'), np.array([i.frame for i in value.inlets_list],
                                                                                 dtype=np.int32)
            yield ValveDataCodec.varname(name, 'inlets_list', 'type'), np.array(
                [value.onlytype.index(i.type) for i in value.inlets_list], dtype=np.int32)
            yield ValveDataCodec.varname(name, 'inlets_list', 'reference', 'ids'), np.array(
                [list(i.reference.id) + [i.reference.nr] for i in value.inlets_list], dtype=np.int32)
            I = len(value.inlets_ids)
            yield ValveDataCodec.varname(name, 'inlets_list', 'reference', 'name'), np.fromiter(
                chain(*[i.reference.name for i in value.inlets_list]), dtype='S1').reshape(I, 3)
            yield ValveDataCodec.varname(name, 'inlets_ids'), np.array(value.inlets_ids, dtype=np.int32)
            yield ValveDataCodec.varname(name, 'clusters'), np.array(value.clusters, dtype=np.int32)
            yield ValveDataCodec.varname(name, 'number_of_clustered_inlets'), np.array(
                [value.number_of_clustered_inlets],
                dtype=np.int32)
            yield ValveDataCodec.varname(name, 'spheres'), np.array(
                [s.center.tolist() + [s.radius] for s in value.spheres])
            yield ValveDataCodec.varname(name, 'spheres', 'nr'), np.array([s.nr for s in value.spheres], dtype=np.int32)
            yield ValveDataCodec.varname(name, 'passing'), np.array([value.passing], dtype='i1')
            yield ValveDataCodec.varname(name, 'tree'), np.fromiter(repr(value.tree), dtype='S1')

    @staticmethod
    def decode(name, data):
        if name == 'center_of_system':
            return data[name][:].copy()
        if name == 'all_res':
            # create empty all_res object
            # read layers
            layers = data[ValveDataCodec.varname(name, 'layers')][:].copy()
            return ResidueSelection(((l, data[ValveDataCodec.varname(name, 'layer', l)]) for l in layers))
        if name == 'number_frame_rid_in_object':
            out = []
            # number of layers
            layers = int(data[ValveDataCodec.varname(name, 'layers')][:].copy())
            for nr in range(layers):
                out_ = []
                sizes = data[ValveDataCodec.varname(name, 'layer', nr, 'sizes')]
                layer = data[ValveDataCodec.varname(name, 'layer', nr)]
                seek = 0
                for s in sizes:
                    out_.append(layer[seek:seek + s].copy())
                    seek += s
                out.append(out_)
            return out
        if name == 'paths':
            out = []
            layers = data[ValveDataCodec.varname(name, 'layers')]
            for N in layers:
                names = data[ValveDataCodec.varname(name, 'layer', N, 'names')]
                ids = data[ValveDataCodec.varname(name, 'layer', N, 'ids')]
                mmf = data[ValveDataCodec.varname(name, 'layer', N, 'min_max_frames')]
                object_sizes = data[ValveDataCodec.varname(name, 'layer', N, 'object', 'sizes')]
                object_frames = data[ValveDataCodec.varname(name, 'layer', N, 'object')]
                scope_sizes = data[ValveDataCodec.varname(name, 'layer', N, 'scope', 'sizes')]
                scope_frames = data[ValveDataCodec.varname(name, 'layer', N, 'scope')]
                seek_object = 0
                seek_scope = 0
                for osize, ssize, n, pid in zip(object_sizes, scope_sizes, names, ids):
                    out.append(
                        GenericPaths(tuple(map(int, (N, pid))), name_of_res=str(n.tostring()), min_pf=int(mmf[0]),
                                     max_pf=int(mmf[1])))
                    foo = list(SmartRange(fast_minc_seq=object_frames[seek_object:seek_object + osize]).get())
                    seek_object += osize
                    fos = list(SmartRange(fast_minc_seq=scope_frames[seek_scope:seek_scope + ssize]).get())
                    seek_scope += ssize
                    out[-1].add_foos(foo, fos)
            return out
        if name == 'spaths':
            out = []
            layers = data[ValveDataCodec.varname(name, 'layers')]
            for N in layers:
                names = data[ValveDataCodec.varname(name, 'layer', N, 'names')]
                ids = data[ValveDataCodec.varname(name, 'layer', N, 'ids')]
                is_single = data[ValveDataCodec.varname(name, 'layer', N, 'single')]
                frames_table = data[ValveDataCodec.varname(name, 'layer', N, 'frames')]
                object_strict = data[ValveDataCodec.varname(name, 'layer', N, 'object')]
                single_path_nr = 0
                seek_object = 0
                for n, pid, ft, iss in zip(names, ids, frames_table, is_single):
                    spid = SinglePathID(path_id=tuple(map(int, (N, pid[0]))), nr=int(pid[-1]), name=str(n.tostring()))
                    path = list(range(ft[0], ft[1] + 1))
                    if iss:
                        out.append(SinglePath(spid, [[], [], []], [[], [], []]))
                        seek = 0
                        path_in = path[seek:seek + ft[2]]
                        seek += ft[2]
                        path_object = path[seek:seek + ft[3]]
                        seek += ft[3]
                        path_out = path[seek:seek + ft[4]]
                        seek += ft[4]
                        path_object_strict_size = data[ValveDataCodec.varname(name, 'layer', N, 'object', 'sizes')][
                            single_path_nr]
                        single_path_nr += 1
                        path_object_strict = list(SmartRange(
                            fast_minc_seq=object_strict[seek_object:seek_object + path_object_strict_size]).get())
                        seek_object += path_object_strict_size
                        out[-1].add_paths4(path_in, path_object, path_object_strict, path_out)
                    else:
                        out.append(PassingPath(spid, path, [GenericPathTypeCodes.scope_name] * len(path)))
            return out

        if name == 'ctypes':
            ctypes = data[ValveDataCodec.varname(name)]
            return [InletClusterExtendedType(cc[0], cc[2], cc[3], cc[1]) for cc in ([None if c == -1 else c for c in ct] for ct in ctypes)]

        if name in ['master_paths', 'master_paths_smooth']:
            out = {}
            keys = data[ValveDataCodec.varname(name, 'keys')]
            names = data[ValveDataCodec.varname(name, 'names')]
            ids = data[ValveDataCodec.varname(name, 'ids')]
            frames_table = data[ValveDataCodec.varname(name, 'frames')]
            widths = data[ValveDataCodec.varname(name, 'widths')]
            coords = data[ValveDataCodec.varname(name, 'coords')]  # corrds use the same seek as widths
            object_strict = data[ValveDataCodec.varname(name, 'object')]
            seek_widths = 0
            seek_object = 0
            for mpnr, (k, n, pid, ft) in enumerate(zip(keys, names, ids, frames_table)):
                key = InletClusterGenericType(*[None if c == -1 else c for c in k])
                spid = SinglePathID(path_id=tuple(map(int, pid[:2])), nr=int(pid[-1]), name=str(n.tostring()))
                path = list(range(ft[0], ft[1] + 1))

                mp = SinglePath(spid, [[], [], []], [[], [], []])
                seek = 0
                path_in = path[seek:seek + ft[2]]
                seek += ft[2]
                path_object = path[seek:seek + ft[3]]
                seek += ft[3]
                path_out = path[seek:seek + ft[4]]
                seek += ft[4]

                path_object_strict_size = data[ValveDataCodec.varname(name, 'object', 'sizes')][mpnr]
                path_object_strict = list(
                    SmartRange(fast_minc_seq=object_strict[seek_object:seek_object + path_object_strict_size]).get())
                seek_object += path_object_strict_size

                fsrs = FakeSingleResidueSelection(spid.id, range(ft[0], ft[1] + 1),
                                                  coords[seek_widths:seek_widths + ft[1] - ft[0] + 1].copy())
                mp = SinglePath(spid, [[], [], []], [[], [], []], single_res_selection=fsrs)

                mp.add_paths4(path_in, path_object, path_object_strict, path_out)
                mp = MasterPath(mp)
                mp.add_width(widths[seek_widths:seek_widths + ft[1] - ft[0] + 1].copy())
                seek_widths += (ft[1] - ft[0] + 1)

                out.update({key: mp})
            return out

        if name == 'inls':
            if ValveDataCodec.varname(name, 'center_of_system') in data:
                center_of_system = data[ValveDataCodec.varname(name, 'center_of_system')][:].copy()
            else:
                center_of_system = None
            # onlytype = data[ValveDataCodec.varname(name, 'onlytype')][:].copy()
            # onlytype = json.loads(str(onlytype.tostring()))
            # onlytype = [tuple(ot) for ot in onlytype]
            passing = bool(data[ValveDataCodec.varname(name, 'passing')][:].copy())
            inls = Inlets([], center_of_system=center_of_system, passing=passing, onlytype=onlytype)

            coords = data[ValveDataCodec.varname(name, 'inlets_list', 'coords')]
            frame = data[ValveDataCodec.varname(name, 'inlets_list', 'frame')]
            type_ = data[ValveDataCodec.varname(name, 'inlets_list', 'type')]
            ref_ids = data[ValveDataCodec.varname(name, 'inlets_list', 'reference', 'ids')]
            ref_name = data[ValveDataCodec.varname(name, 'inlets_list', 'reference', 'name')]
            for c, f, t, pid, n in zip(coords, frame, type_, ref_ids, ref_name):
                spid = SinglePathID(path_id=tuple(map(int, pid[:2].copy())), nr=int(pid[-1].copy()),
                                    name=str(n.copy().tostring()))
                # inls.inlets_list.append(Inlet(coords=c.copy(), type=onlytype[t.copy()], reference=spid, frame=f.copy()))
                inls.inlets_list.append(Inlet(coords=c.copy(), reference=spid, frame=f.copy()))

            inls.inlets_ids = data[ValveDataCodec.varname(name, 'inlets_ids')][:].copy().tolist()
            inls.clusters = data[ValveDataCodec.varname(name, 'clusters')][:].copy().tolist()
            inls.number_of_clustered_inlets = int(
                data[ValveDataCodec.varname(name, 'number_of_clustered_inlets')][:].copy())

            spheres = data[ValveDataCodec.varname(name, 'spheres')]
            spheres_nr = data[ValveDataCodec.varname(name, 'spheres', 'nr')]
            for s, nr in zip(spheres, spheres_nr):
                inls.spheres.append(Sphere(s[:3].copy(), s[-1].copy(), nr.copy()))

            inls.tree = SimpleTree(treestr=str(data[ValveDataCodec.varname(name, 'tree')][:].copy().tostring()))

            return inls


class ValveDataAccess_nc(ValveDataAccess):
    not_variable = ['version', 'aquaduct_version', 'ValveDataCodec']

    def open(self):
        if hasattr(netcdf, 'Dataset'):
            # netcdf4
            self.data_file = netcdf.Dataset(self.data_file_name, self.mode)
        else:
            # scipy
            self.data_file = netcdf.netcdf_file(self.data_file_name, self.mode)
        if self.mode == 'w':
            self.set_variable('version', np.array(version(), dtype=np.int16))
            self.set_variable('aquaduct_version', np.array(version(), dtype=np.int16))
            self.set_variable('ValveDataCodec', np.array(ValveDataCodec.version, dtype=np.int16))  # not used right now
        elif self.mode == 'r':
            versions = dict([(k, tuple(v)) for k, v in zip(['version', 'aquaduct_version'],
                                                           (self.get_variable('version'),
                                                            self.get_variable('aquaduct_version')))])
            check_versions(versions)

    def close(self):
        self.data_file.close()

    def get_variable(self, name, copy=True):
        # print name
        if copy:
            return self.data_file.variables[name][:].copy()
        return self.data_file.variables[name]

    def set_variable(self, name, value):
        # print name, value
        assert self.mode == "w"
        # value has to be ndarray
        assert isinstance(value, np.ndarray)
        # create dimensions
        dimensions = []
        for nr, d in enumerate(value.shape):
            dimensions.append("%s%d" % (name, nr))
            self.data_file.createDimension(dimensions[-1], d)
        # create variable
        if hasattr(netcdf, 'Dataset'):
            v = self.data_file.createVariable(name, value.dtype, tuple(dimensions), zlib=True)
        else:
            v = self.data_file.createVariable(name, value.dtype, tuple(dimensions))
        # fill variable
        v[:] = value

    def dump(self, **kwargs):
        for name, value in kwargs.items():
            for nname, vvalue in ValveDataCodec.encode(name, value):
                self.set_variable(nname, vvalue)

    @dictify
    def load(self):
        # names of data objects
        names = list(
            set([name.split('.')[0] for name in list(self.data_file.variables.keys()) if name not in self.not_variable]))
        all_names = [name for name in list(self.data_file.variables.keys()) if name not in self.not_variable]
        for name in names:
            # if 'inls' in name: continue
            # read all parts of object and decode
            this_object = dict(((n, self.get_variable(n, copy=False)) for n in all_names if
                                name == n or (name + '.') == n[:len(name) + 1]))
            yield name, ValveDataCodec.decode(name, this_object)
            for k in list(this_object.keys()):
                v = this_object.pop(k)[:]
                if hasattr(v, 'copy'):
                    v.copy()
                del v


################################################################################

class ValveDataAccess_numpy(ValveDataAccess):
    pass


################################################################################

class ValveDataAccess_pickle(ValveDataAccess):
    mimic_old_var_name = 'aq_data_to_save'
    unknown_names = 'UNK'

    def open(self):
        # open file
        self.data_file = gzip.open(self.data_file_name, mode=self.mode, compresslevel=9)
        # if mode is w save header with version etc
        if self.mode == 'w':
            pickle.dump({'version': version(),
                         'aquaduct_version': version()}, self.data_file)
        elif self.mode == 'r':
            versions = pickle.load(self.data_file)
            check_versions(versions)
            # loaded data!
            self.data = {}
            try:
                while True:
                    loaded_data = pickle.load(self.data_file)
                    self.data.update(loaded_data)
            except:  # TODO: remove it!
                pass

    def close(self):
        self.data_file.close()

    def load(self):
        data = {}
        # this is to mimic v0.3 behaviour
        for _name, _value in self.data.items():
            if _name == self.mimic_old_var_name:
                for name, value in _value.items():
                    '''
                    if isinstance(value, CompactSelectionMDA):
                        value = value.toSelectionMDA(self.reader)
                        # with self.reader.get() as traj_reader:
                        #    value = value.toSelectionMDA(traj_reader)
                    '''
                    # TODO: following is to overcome problems with missing names when data is <0.4
                    ################################################################################
                    if name == 'paths':
                        for path_name, path in value.items():
                            if not hasattr(path, 'name'):
                                path.name = self.unknown_names
                    if name == 'spaths':
                        for spath in value:
                            if not hasattr(spath.id, 'name'):
                                spath.id.name = self.unknown_names
                    if name == 'inls':
                        if not hasattr(value, 'spheres'):
                            value.spheres = []
                        if hasattr(value, 'refs'):
                            for r in value.refs:
                                if not hasattr(r, 'name'):
                                    r.name = self.unknown_names
                    ################################################################################
                    data.update({name: value})
                break
            else:
                for name, value in self.data.items():
                    '''
                    if isinstance(value, CompactSelectionMDA):
                        value = value.toSelectionMDA(self.reader)
                        # with self.reader.get() as traj_reader:
                        #    value = value.toSelectionMDA(traj_reader)
                    '''
                    # TODO: following is to overcome problems with missing names when data is <0.4
                    ################################################################################
                    '''
                    if name == 'paths':
                        for path_name,path in value.iteritems():
                            if not hasattr(path,'name'):
                                path.name = self.unknown_names
                    if name == 'spaths':
                        for spath in value:
                            if not hasattr(spath.id,'name'):
                                spath.id.name = self.unknown_names
                    if name == 'inls':
                        if not hasattr(value,'spheres'):
                            value.spheres = []
                        if hasattr(value,'refs'):
                            for r in value.refs:
                                if not hasattr(r, 'name'):
                                    r.name = self.unknown_names
                    '''
                    ################################################################################
                    data.update({name: value})
                break
        return data

    def dump(self, **kwargs):
        for name, value in kwargs.items():
            self.set_variable(name, value)

    def get_variable(self, name):
        assert self.mode == "r"
        value = self.data[name]

        return value

    def set_variable(self, name, value):
        assert self.mode == "w"
        '''
        if isinstance(value, SelectionMDA):
            value = CompactSelectionMDA(value)
        '''
        # options are passed as dictionaries already
        '''
        if 'options' in name:
            value = value._asdict()
        '''
        if False:  # hasattr(value,'simple_dump'):
            pickle.dump({name: value.simple_dump()}, self.data_file)
        else:
            pickle.dump({name: value}, self.data_file)



# -*- coding: utf-8 -*-

# Aqua-Duct, a tool facilitating analysis of the flow of solvent molecules in molecular dynamic simulations
# Copyright (C) 2016-2018  Tomasz Magdziarz, Alicja Płuciennik, Michał Stolarczyk <info@aquaduct.pl>
# Copyright (C) 2019  Tomasz Magdziarz <info@aquaduct.pl>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
import scipy.linalg


class NullPrepocess(object):
    def __init__(self):
        pass

    def build(self, X):
        pass

    def __call__(self, X):
        return X

    def undo(self, X):
        return X


class Center(object):
    def __init__(self):
        pass

    def build(self, X):
        self.mean = np.mean(X, 0)

    def __call__(self, X):
        return X - self.mean

    def undo(self, X):
        return X + self.mean


class Normalize(object):
    def __init__(self):
        pass

    def build(self, X):
        self.std = np.std(X, 0)

    def __call__(self, X):
        return X / self.std

    def undo(self, X):
        return X * self.std


class Standartize(object):
    def __init__(self):
        self.center = Center()
        self.normalize = Normalize()

        test_arr = np.random.random((10, 3))
        test_cntr = np.random.random((1, 3))
        P = Polarize(center=test_cntr)
        P.build(test_arr)
        test_arr_p = P(test_arr)
        Pp = P.undo(test_arr_p) - test_arr
        [self.assertAlmostEqual(x, 0, 7) for x in Pp.ravel()]

    def test_polarize_in(self):
        P = Polarize(center=np.random.random((1, 3)))
        self.assertRaises(TypeError, P, 'cupkaces')
        self.assertRaises(TypeError, Polarize, 'cupcakes')
        self.assertRaises(TypeError, P, np.random.random((1, 2)))


if __name__ == '__main__':
    def build(self, X):
        self.center.build(X)
        self.normalize.build(X)


    def __call__(self, X):
        return self.normalize(self.center(X))


    def undo(self, X):
        return self.center.undo(self.normalize.undo(X))


class Polarize(object):
    # TODO: equal variance of t and f
    def __init__(self, center=np.array([0, 0, 0]), rvar=0.1, equaltf=True):
        '''
        Prepocessing filter for 3D cartesian coordinates transformation to spherical coordinates.

        .. note::

            Component f is in range 0 - 2\pi.

        :param center: Center of the hypothetical sphere.
        :param rvar: Desired amount of variance of *r* component measured as fraction of mean *t* and *f* variance.
        :param equaltf: If set ``True``, *t* range is scaled to *f*.
        '''

        if not isinstance(center, np.ndarray):
            raise TypeError('Constructor called with center param of invalid type')
        elif np.shape(center) != (3,):
            raise TypeError('Constructor called with center param of invalid shape')
        else:
            self.center = center

        self.rvar = rvar
        self.equaltf = equaltf
        self.tmean = 0
        self.fmean = 0
        self.rvar_factor = 1

    @property
    def _mt(self):
        return 2. if self.equaltf else 1.

    def _Xrtf(self, X):
        X = X - self.center
        # ISO, f + pi
        r = (X ** 2).sum(1) ** 0.5
        t = np.arccos(X[:, 2] / r)  # 0,pi
        f = np.arctan2(X[:, 1], X[:, 0]) + np.pi  # -pi,pi # f+pi 0,2pi
        return X, r, t, f

    def _circle_tf(self, t, f):
        return t % np.pi, f % (2 * np.pi)

    def build(self, X):
        X, r, t, f = self._Xrtf(X)
        # calculate mean values for polar components (circle it)
        self.tmean, self.fmean = self._circle_tf(np.pi / 2 - np.mean(t), np.pi - np.mean(f))
        tf_var = (np.var(t) * 2 + np.var(f)) / 4.
        if np.std(r) > 0:
            self.rvar_factor = ((tf_var * self.rvar) ** 0.5) * (1. / np.std(r))
        else:
            self.rvar_factor = 1.

    def __call__(self, X):
        X, r, t, f = self._Xrtf(X)
        # center t and f with appropriate mean values
        t += self.tmean
        f += self.fmean
        # take care of polarity
        t, f = self._circle_tf(t, f)
        return np.array([r * self.rvar_factor, t * self._mt, f]).T

    def undo(self, X):
        # remove centering
        t = X[:, 1] / self._mt - self.tmean
        f = X[:, 2] - self.fmean
        # take care of polarity
        t, f = self._circle_tf(t, f)
        # correct f to iso
        f -= np.pi
        # get xyz
        x = X[:, 0] / self.rvar_factor * np.sin(t) * np.cos(f)
        y = X[:, 0] / self.rvar_factor * np.sin(t) * np.sin(f)
        z = X[:, 0] / self.rvar_factor * np.cos(t)
        return np.array([x, y, z]).T + self.center


class PCA(object):
    def __init__(self, preprocess=NullPrepocess()):
        self.preprocess = preprocess

    def build(self, X):
        self.preprocess.build(X)
        # SVD
        self.U, self.d, self.Pt = scipy.linalg.svd(self.preprocess(X), full_matrices=False)
        assert np.all(self.d[:-1] >= self.d[1:]), "SVD error."  # sorted
        # PCA
        self.T = self.U * self.d
        self.eigen = self.d ** 2
        self.sumvariance = np.cumsum(self.eigen)
        self.sumvariance /= self.sumvariance[-1]

    @property
    def P(self):
        return self.Pt.T

    def __call__(self, X, pc=None):
        if pc:
            return np.dot(self.preprocess(X), self.P[:, pc])
        return np.dot(self.preprocess(X), self.P)

    def undo(self, T, pc=None):
        if pc:
            return self.preprocess.undo(np.dot(T, self.Pt[pc, :]))
        return self.preprocess.undo(np.dot(T, self.Pt))



# -*- coding: utf-8 -*-

# This program is distributed in the hope that it will be useful,
# Aqua-Duct, a tool facilitating analysis of the flow of solvent molecules in molecular dynamic simulations
# Copyright (C) 2016-2018  Tomasz Magdziarz, Alicja Płuciennik, Michał Stolarczyk <info@aquaduct.pl>
# Copyright (C) 2019  Tomasz Magdziarz, Michał Banas <info@aquaduct.pl>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


# Module contains mostly stuff need for stage V

from functools import wraps

import numpy as np
from aquaduct.geom import traces
from aquaduct.traj.paths import PassingPath


class SpathIdHeader(object):
    name = 'ID'
    format = '%9s'

    def __call__(self):
        return [self.name], [self.format]


'''
def spath_id_header():
    return ['ID'], ['%9s']
'''
spath_id_header = SpathIdHeader()


def spath_name_header():
    return ['RES'], ['%4s']


def add_path_id_head(gen):
    @wraps(gen)
    def patched(*args, **kwargs):
        sph, splt = list(zip(spath_id_header(), spath_name_header()))
        sph = [e[0] for e in sph]
        splt = [e[0] for e in splt]

        add_id = True
        if 'add_id' in kwargs:
            add_id = kwargs.pop('add_id')
        h, lt = gen(*args, **kwargs)
        if add_id:
            return sph + h, splt + lt
        return h, lt

    return patched


def add_path_id(gen):
    @wraps(gen)
    def patched(spath, *args, **kwargs):
        add_id = True
        if 'add_id' in kwargs:
            add_id = kwargs.pop('add_id')
        line = gen(spath, *args, **kwargs)
        if add_id:
            line = [spath.id, spath.id.name] + line
        return line

    return patched


def size_header():
    return ['Size', 'Size%'], ['%7d', '%6.2f']


def add_size_head(gen):
    sph, splt = size_header()

    @wraps(gen)
    def patched(*args, **kwargs):
        h, lt = gen(*args, **kwargs)
        return sph + h, splt + lt

    return patched


def add_size(gen):
    @wraps(gen)
    def patched(spaths, add_size=True, add_size_p100=None, *args, **kwargs):
        line = gen(spaths, *args, **kwargs)
        if add_size_p100 is not None:
            line = [len(spaths) / float(add_size_p100) * 100] + line
        if add_size:
            line = [len(spaths)] + line
        return line

    return patched


def ctype_id_header():
    return ['CType'], ['%7s']


def add_ctype_id_head(gen):
    sph, splt = ctype_id_header()

    @wraps(gen)
    def patched(*args, **kwargs):
        h, lt = gen(*args, **kwargs)
        return sph + h, splt + lt

    return patched


def add_ctype_id(gen):
    @wraps(gen)
    def patched(ctype, something, add_id=True, *args, **kwargs):
        line = gen(ctype, something, *args, **kwargs)
        if add_id:
            line = [str(ctype)] + line
        return line

    return patched


@add_path_id_head
def spath_basic_info_header():
    header = 'BeginF InpF ObjF ObjFS OutF EndF'.split()
    line_template = ['%7d'] * len(header)
    return header, line_template


@add_path_id
def spath_basic_info(spath):
    line = [spath.begins]
    if not isinstance(spath, PassingPath):
        line.extend(list(map(len, (spath.path_in, spath.path_object))))
        line.append(spath.path_object_strict_len)
        line.append(len(spath.path_out))
        # line.extend(map(len, (spath.path_in, spath.path_object, spath.path_out)))
    else:
        line += [float('nan')] * 4
    line.append(spath.ends)
    return line


@add_path_id_head
def spath_lenght_total_info_header(total=None):
    header = 'InpL ObjL OutL'.split()
    if total:
        header = ['TotL'] + header
    line_template = ['%9.1f'] * len(header)
    return header, line_template


@add_path_id
def spath_lenght_total_info(spath, totalonly=False, total=False):
    # total and totalonly are internal flags
    # to calculate: total len, in len, obj len, out len call:
    # total=True (this will calculate total)
    # to skip calculation of total call:
    # total=False, totalonly=False (this will calculate in, obj, and out lens)
    # to calculate total len call:
    # total=False, totalonly=True
    line = []
    if not total:
        if not totalonly:
            for t in traces.midpoints(spath.coords):
                if len(t) > 1:  # traces.length_step_std requires at least 2 points
                    line.append(traces.length_step_std(t)[0])
                else:
                    line.append(float('nan'))
            return line
        else:
            for t in traces.midpoints((spath.coords_cont,)):
                if len(t) > 1:  # traces.length_step_std requires at least 2 points
                    line.append(traces.length_step_std(t)[0])
                else:
                    line.append(float('nan'))
            return line
    line += spath_lenght_total_info(spath, add_id=False, total=False, totalonly=True)
    if not isinstance(spath, PassingPath):
        line += spath_lenght_total_info(spath, add_id=False, total=False, totalonly=False)
    else:
        line += [float('nan')] * 3
    return line


@add_path_id
def spath_frames_total_info(spath, totalonly=False, total=False):
    # total and totalonly are internal flags
    # to calculate: total len, in len, obj len, out len call:
    # total=True (this will calculate total)
    # to skip calculation of total call:
    # total=False, totalonly=False (this will calculate in, obj, and out lens)
    # to calculate total len call:
    # total=False, totalonly=True
    line = []
    if not total:
        if not totalonly:
            for t in spath.coords:
                line.append(len(t))
            return line
        else:
            for t in (spath.coords_cont,):
                line.append(len(t))
            return line
    line += spath_frames_total_info(spath, add_id=False, total=False, totalonly=True)
    if not isinstance(spath, PassingPath):
        line += spath_frames_total_info(spath, add_id=False, total=False, totalonly=False)
    else:
        line += [float('nan')] * 3
    return line


@add_path_id_head
def spath_steps_info_header(total=None):
    header = 'InpS InpStdS ObjS ObjStdS OutS OutStdS'.split()
    if total:
        header = ['TotS', 'TotStdS'] + header
    line_template = ['%8.2f', '%8.3f'] * int(len(header) / 2)
    return header, line_template


@add_path_id
def spath_steps_info(spath, total=None):
    line = []
    if not total:
        if not isinstance(spath, PassingPath):
            for t in traces.midpoints(spath.coords):
                if len(t) > 0:
                    line.extend(traces.length_step_std(t)[1:])
                else:
                    line.extend([float('nan'), float('nan')])
        else:
            line += [float('nan'), float('nan')] * 3
        return line
    line += spath_steps_info(spath, add_id=False, total=False)
    t = next(traces.midpoints((spath.coords_cont,)))
    if len(t) > 0:
        line = list(traces.length_step_std(t)[1:]) + line
    else:
        line = [float('nan'), float('nan')] + line
    return line


@add_path_id
def spath_frames_info(spath, total=None):
    line = []
    if not total:
        for t in spath.coords:
            line.append(len(t))
        return line
    line += spath_steps_info(spath, add_id=False, total=False)
    if not isinstance(spath, PassingPath):
        t = spath.coords_cont
        line = [len(t)] + line
    else:
        line = [float('nan')] + line
    return line


@add_path_id_head
def spath_ctype_header():
    header, line_template = ctype_id_header()
    return header, line_template


@add_path_id
def spath_ctype(spath, ctype=None):
    line = [str(ctype)]
    return line


@add_path_id_head
def spath_full_info_header(total=None):
    header = []
    line_template = []
    for h, lt in (spath_basic_info_header(add_id=False),
                  spath_lenght_total_info_header(add_id=False, total=total),
                  spath_steps_info_header(add_id=False, total=total),
                  spath_ctype_header(add_id=False)):
        header += h
        line_template += lt
    return header, line_template


@add_path_id
def spath_full_info(spath, ctype=None, total=None):
    line = []
    for l in (spath_basic_info(spath, add_id=False),
              spath_lenght_total_info(spath, add_id=False, total=total),
              spath_steps_info(spath, add_id=False, total=total),
              spath_ctype(spath, ctype=ctype, add_id=False)):
        line += l
    return line


@add_size_head
def spaths_lenght_total_header():
    header = 'Tot TotStd Inp InpStd Obj ObjStd Out OutStd'.split()
    line_template = ['%9.1f', '%9.2f'] * int(len(header) / 2)
    return header, line_template


@add_size
def spaths_length_total(spaths):
    line = []
    d4s = []
    for sp in spaths:
        if not isinstance(sp, PassingPath):
            d4s.append(spath_lenght_total_info(sp, add_id=False))
    d4s = np.array(d4s)
    if d4s.size:
        line.extend(np.mean(d4s, 0))
        line.extend(np.std(d4s, 0))
    else:
        line = [float('nan')] * 6
    # total
    d4s = []
    for sp in spaths:
        d4s.append(spath_lenght_total_info(sp, totalonly=True, add_id=False))
    d4s = np.array(d4s)
    if d4s.size:
        line.extend(np.mean(d4s, 0))
        line.extend(np.std(d4s, 0))
    else:
        line += [float('nan')] * 4

    return [line[6], line[7], line[0], line[3], line[1], line[4], line[2], line[5]]


@add_size
def spaths_frames_total(spaths):
    line = []
    d4s = []
    for sp in spaths:
        if not isinstance(sp, PassingPath):
            d4s.append(spath_frames_total_info(sp, add_id=False))
    d4s = np.array(d4s)
    if d4s.size:
        line.extend(np.mean(d4s, 0))
        line.extend(np.std(d4s, 0))
    else:
        line = [float('nan')] * 6
    # total
    d4s = []
    for sp in spaths:
        d4s.append(spath_frames_total_info(sp, totalonly=True, add_id=False))
    d4s = np.array(d4s)
    if d4s.size:
        line.extend(np.mean(d4s, 0))
        line.extend(np.std(d4s, 0))
    else:
        line += [float('nan')] * 4

    return [line[6], line[7], line[0], line[3], line[1], line[4], line[2], line[5]]


@add_ctype_id_head
def ctypes_spaths_info_header():
    header, line_template = spaths_lenght_total_header()
    return header, line_template


@add_ctype_id
def ctypes_spaths_info(ctype, spaths, show='len', add_size_p100=None):
    # show could be len or frames
    line = []
    if show == 'len':
        line += spaths_length_total(spaths, add_size_p100=add_size_p100)
    if show == 'frames':
        line += spaths_frames_total(spaths, add_size_p100=add_size_p100)
    return line


def plot_spaths_traces(spaths, spp=None, name=None, split=False, states=False, separate=False, smooth=None):
    if states or separate:
        spaths_iter = spaths
    else:
        spaths_iter = [spaths]
    state = None
    name_separate = ''
    for nr, sp in enumerate(spaths_iter):
        if states:
            state = nr + 1
        if separate:
            name_separate = '_%d' % (nr + 1)
        if states or separate:
            sp = [sp]
        if split:
            # FIXME: if path(s) is empty this will probably produce CGO_END without CGO_BEGIN
            spp.paths_trace(sp, name=name + '_in' + name_separate, plot_walk=False, plot_object=False, plot_out=False,
                            state=state, smooth=smooth)
            spp.paths_trace(sp, name=name + '_obj' + name_separate, plot_walk=False, plot_in=False, plot_out=False,
                            state=state, smooth=smooth)
            spp.paths_trace(sp, name=name + '_out' + name_separate, plot_walk=False, plot_in=False, plot_object=False,
                            state=state, smooth=smooth)
            spp.paths_trace(sp, name=name + '_walk' + name_separate, plot_in=False, plot_object=False, plot_out=False,
                            state=state, smooth=smooth)
        else:
            spp.paths_trace(sp, name=name + name_separate, state=state, smooth=smooth)


def plot_spaths_inlets(spaths, spp=None, name=None, states=False, separate=False, smooth=None):
    if states or separate:
        spaths_iter = spaths
    else:
        spaths_iter = [spaths]
    state = None
    name_separate = ''
    for nr, sp in enumerate(spaths_iter):
        if states:
            state = nr + 1
        if separate:
            name_separate = '_%d' % (nr + 1)
        if states or separate:
            sp = [sp]
        spp.paths_inlets(sp, name=name + name_separate, state=state, smooth=smooth)



# -*- coding: utf-8 -*-

# Aqua-Duct, a tool facilitating analysis of the flow of solvent molecules in molecular dynamic simulations
# Copyright (C) 2016-2018  Tomasz Magdziarz, Alicja Płuciennik, Michał Stolarczyk <info@aquaduct.pl>
# Copyright (C) 2019  Tomasz Magdziarz <info@aquaduct.pl>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from aquaduct import logger

from collections import namedtuple

import numpy as np
from scipy.spatial.distance import cdist
import copy

from aquaduct.utils import clui
from aquaduct.utils.helpers import lind
from aquaduct import __mail__


class Sphere(namedtuple('Sphere', 'center radius nr')):
    """
    Simple sphere class.
    """

    def is_point_within(self, point):
        return self.radius > cdist(np.matrix(self.center), np.matrix(point), metric='euclidean')

    def is_sphere_within(self, sphere):
        center, radius = sphere
        return self.radius > cdist(np.matrix(self.center), np.matrix(center), metric='euclidean') + radius

    def is_sphere_cloud(self, sphere):
        center, radius = sphere
        return self.radius > cdist(np.matrix(self.center), np.matrix(center), metric='euclidean') - radius


def do_cut_thyself(spheres_passed, progress=False):
    # returns noredundant spheres
    # make a deep copy?
    # TODO: this is not memory efficient?
    spheres = copy.copy(spheres_passed)
    N = len(spheres)
    if progress:
        clui.message("Barber, cut thyself:")
        pbar = clui.pbar(N)
    noredundat_spheres_count = 0
    redundat_spheres = []
    while True:
        spheres.sort(key=lambda s: s.radius, reverse=True)
        spheres_coords = np.array([sphe.center for sphe in spheres])
        spheres_radii = np.array([sphe.radius for sphe in spheres])
        noredundat_spheres = []
        while spheres:
            big = spheres.pop(0)
            center, radius, nr = big
            distances = cdist(np.matrix(center), spheres_coords[1:], metric='euclidean').flatten()
            # add radii
            distances += spheres_radii[1:]
            # remove if distance <= radius
            to_keep = distances > radius
            to_remove = ~to_keep
            # do keep
            spheres_coords = spheres_coords[1:][to_keep]
            spheres_radii = spheres_radii[1:][to_keep]
            # do keep spheres
            to_keep_ids = np.argwhere(to_keep).flatten().tolist()
            to_remove_ids = np.argwhere(to_remove).flatten().tolist()
            redundat_spheres.extend(lind(spheres, to_remove_ids))
            spheres = lind(spheres, to_keep_ids)
            # add big to non redundant
            noredundat_spheres.append(big)
            if progress:
                # pbar.update(N - len(spheres))
                pbar.update(len(redundat_spheres))
            logger.debug("Removal of redundant cutting places: done %d, to analyze %d" % (
                len(noredundat_spheres), len(spheres)))
        if len(noredundat_spheres) == noredundat_spheres_count:
            logger.debug("Removal of redundant cutting places done. %d non redundant spheres found." % len(
                noredundat_spheres))
            break
        else:
            noredundat_spheres_count = len(noredundat_spheres)
            spheres = noredundat_spheres
    if progress:
        pbar.finish()

    assert len(noredundat_spheres) + len(
        redundat_spheres) == N, "Inconsistent number of not and redundant spheres. Please send a bug report to the developer(s): %s" % __mail__
    # sorting
    noredundat_spheres.sort(key=lambda s: s.nr)
    redundat_spheres.sort(key=lambda s: s.nr)
    return noredundat_spheres, redundat_spheres



# -*- coding: utf-8 -*-

# This program is distributed in the hope that it will be useful,
# Aqua-Duct, a tool facilitating analysis of the flow of solvent molecules in molecular dynamic simulations
# Copyright (C) 2016-2018  Tomasz Magdziarz, Alicja Płuciennik, Michał Stolarczyk <info@aquaduct.pl>
# Copyright (C) 2019  Tomasz Magdziarz, Michał Banas <info@aquaduct.pl>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import operator
import os
import re
import sys
import numpy as np

from aquaduct.apps.data import CRIC
from aquaduct import logger
from aquaduct.apps.valve import asep
from aquaduct.geom import traces
from aquaduct.geom.cluster_available_methods import AVAILABLE_METHODS as available_clustering_methods
from aquaduct.geom.cluster import BarberCluster, PerformClustering
from aquaduct.geom.smooth import WindowSmooth, ActiveWindowSmooth, DistanceWindowSmooth, MaxStepSmooth, \
    WindowOverMaxStepSmooth, ActiveWindowOverMaxStepSmooth, DistanceWindowOverMaxStepSmooth, SavgolSmooth
from sklearn.cluster import DBSCAN, AffinityPropagation, KMeans, MeanShift, Birch
from aquaduct.traj.barber import WhereToCut
from aquaduct.utils import clui
from aquaduct.utils.helpers import is_number, Auto


class NP(object):
    def __init__(self, pbar, next_len=False):
        self.paths = list()
        self.pbar = pbar
        self.next_len = next_len

    def reinit(self, pbar, next_len=False):
        self.pbar = pbar
        self.next_len = next_len

    def next(self, n):
        if n > 1 and self.next_len:
            self.pbar.next(step=n)
        else:
            self.pbar.next()

    def callback_cric_next(self, result):
        CRIC.update_cric(result[-1])
        self.paths.extend(result[:-1])
        self.next(len(result) - 1)

    def callback_next(self, result):
        self.paths.extend(result)
        self.next(len(result))

    def callback_append_next(self, result):
        self.paths.append(result)
        self.pbar.next()


def get_res_in_scope(is_res_in_scope, res):
    res_new = None
    for iris, r in zip(is_res_in_scope, res.iterate_over_residues()):
        if iris:
            if res_new is None:
                res_new = r
            else:
                res_new += r
    return res_new


def get_smooth_method(soptions):
    assert soptions.method in ['window', 'mss', 'window_mss', 'awin',
                               'awin_mss', 'dwin', 'dwin_mss',
                               'savgol'], 'Unknown smoothing method %s.' % soptions.method

    opts = {}
    if 'recursive' in soptions._asdict():
        opts.update({'recursive': int(soptions.recursive)})

    def window_opts():
        if 'window' in soptions._asdict():
            opts.update({'window': int(float(soptions.window))})
        function_opts()

    def awin_dwin_opts():
        if 'window' in soptions._asdict():
            opts.update({'window': float(soptions.window)})
        function_opts()

    def function_opts():
        if 'function' in soptions._asdict():
            assert soptions.function in ['mean', 'median'], 'Unknown smoothing function %s.' % soptions.function
            if soptions.function == 'mean':
                opts.update({'function': np.mean})
            if soptions.function == 'median':
                opts.update({'function': np.median})

    def mss_opts():
        if 'step' in soptions._asdict():
            opts.update({'step': float(soptions.step)})

    def savgol_opts():
        window_length = 5  # TODO: magic constant (default value)
        if 'window' in soptions._asdict():
            window_length = int(float(soptions.window))
            assert window_length % 2 == 1, 'Window in Savgol method must be positive odd number, %d given instead.' % window_length
            opts.update({'window_length': window_length})
        polyorder = 2  # TODO: magic constant (default value)
        if 'polyorder' in soptions._asdict():
            polyorder = int(float(soptions.polyorder))
            assert polyorder > 0, 'Polynomial order should be greater then 0, %d given instead.' % polyorder
            opts.update({'polyorder': polyorder})
        assert polyorder < window_length, 'Polynomial order (%d) should be less then window (%d).' % (
            polyorder, window_length)
        '''
        if 'deriv' in soptions._asdict():
            deriv = int(float(soptions.deriv))
            assert deriv >= 0 and deriv <= polyorder, 'Order of derrivative should be integer greater or equal 0 and less then or equal to Polynomial order (%d), %d given instead.' % (polyorder,deriv)
            opts.update({'deriv': deriv})
        if 'delta' in soptions._asdict():
            delta = float(soptions.delta)
            assert delta >= 0, 'Delta should be greater or equal 0, %f given instead.' % delta
            if 'deriv' in soptions._asdict():
                if deriv == 0:
                    logger.warning('Delta %f make no sense if deriv is 0.' % delta)
            opts.update({'delta': delta})
        mode = 'interp' # TODO: magic constant (default value)
        if 'mode' in soptions._asdict():
            mode = str(soptions.mode)
            assert mode in ['mirror', 'constant', 'nearest', 'wrap', 'interp'], 'Unknown mode %s.' % mode
            opts.update({'mode': mode})
        if 'cval' in soptions._asdict():
            cval = float(soptions.cval)
            if mode != 'constant':
                logger.warning('Cval make no sense if mode is %s.' % mode)
            opts.update({'cval': cval})
        '''

    if soptions.method == 'window':
        window_opts()
        smooth = WindowSmooth(**opts)
    elif soptions.method == 'awin':
        awin_dwin_opts()
        smooth = ActiveWindowSmooth(**opts)
    elif soptions.method == 'dwin':
        awin_dwin_opts()
        smooth = DistanceWindowSmooth(**opts)
    elif soptions.method == 'mss':
        mss_opts()
        smooth = MaxStepSmooth(**opts)
    elif soptions.method == 'window_mss':
        window_opts()
        mss_opts()
        smooth = WindowOverMaxStepSmooth(**opts)
    elif soptions.method == 'awin_mss':
        awin_dwin_opts()
        mss_opts()
        smooth = ActiveWindowOverMaxStepSmooth(**opts)
    elif soptions.method == 'dwin_mss':
        awin_dwin_opts()
        mss_opts()
        smooth = DistanceWindowOverMaxStepSmooth(**opts)
    elif soptions.method == 'savgol':
        savgol_opts()
        smooth = SavgolSmooth(**opts)

    return smooth


def get_auto_barber_options(abo):
    opts = {}
    if 'auto_barber' in abo._asdict():
        opts.update({'selection': str(abo.auto_barber)})
    if 'auto_barber_maxcut' in abo._asdict():
        if is_number(abo.auto_barber_maxcut):
            opts.update({'maxcut': float(abo.auto_barber_maxcut)})
        else:
            opts.update({'maxcut': None})
    if 'auto_barber_mincut' in abo._asdict():
        if is_number(abo.auto_barber_mincut):
            opts.update({'mincut': float(abo.auto_barber_mincut)})
        else:
            opts.update({'mincut': None})
    if 'auto_barber_mincut_level' in abo._asdict():
        opts.update({'mincut_level': bool(abo.auto_barber_mincut_level)})
    if 'auto_barber_maxcut_level' in abo._asdict():
        opts.update({'maxcut_level': bool(abo.auto_barber_maxcut_level)})
    if 'auto_barber_tovdw' in abo._asdict():
        opts.update({'tovdw': bool(abo.auto_barber_tovdw)})
    return opts


def get_clustering_method(coptions, config):
    assert coptions.method in available_clustering_methods, 'Unknown clustering method %s.' % coptions.method

    opts = {}

    def dbscan_opts():
        if 'eps' in coptions._asdict():
            opts.update({'eps': float(coptions.eps)})
        if 'min_samples' in coptions._asdict():
            opts.update({'min_samples': int(coptions.min_samples)})
        if 'metric' in coptions._asdict():
            assert coptions.metric in ['cityblock', 'cosine', 'euclidean',
                                       'manhattan'], "Unknown metric <%s>." % coptions.metric
            opts.update({'metric': str(coptions.metric)})
        if 'algorithm' in coptions._asdict():
            assert coptions.algorithm in ['auto', 'ball_tree', 'kd_tree',
                                          'brute'], "Unknown NN algorithm <%s>." % coptions.algorithm
            opts.update({'algorithm': str(coptions.algorithm)})
        if 'leaf_size' in coptions._asdict():
            opts.update({'leaf_size': int(coptions.leaf_size)})

    def affprop_opts():
        if 'damping' in coptions._asdict():
            opts.update({'damping': float(coptions.damping)})
        if 'convergence_iter' in coptions._asdict():
            opts.update({'convergence_iter': int(coptions.convergence_iter)})
        if 'max_iter' in coptions._asdict():
            opts.update({'max_iter': int(coptions.max_iter)})
        if 'preference' in coptions._asdict():
            opts.update({'preference': float(coptions.preference)})

    def kmeans_opts():
        if 'n_clusters' in coptions._asdict():
            opts.update({'n_clusters': int(coptions.n_clusters)})
        if 'max_iter' in coptions._asdict():
            opts.update({'max_iter': int(coptions.max_iter)})
        if 'n_init' in coptions._asdict():
            opts.update({'n_init': int(coptions.n_init)})
        if 'init' in coptions._asdict():
            assert coptions.init in ['k-means++', 'random'], "Unknown initialization method <%s>." % coptions.init
            opts.update({'init': str(coptions.init)})
        if 'tol' in coptions._asdict():
            opts.update({'tol': float(coptions.tol)})

    def meanshift_opts():
        if 'cluster_all' in coptions._asdict():
            opts.update({'cluster_all': bool(coptions.cluster_all)})
        if 'bandwidth' in coptions._asdict():
            if coptions.bandwidth in (Auto, None):
                opts.update({'bandwidth': coptions.bandwidth})
            else:
                opts.update({'bandwidth': float(coptions.bandwidth)})
        if 'bin_seeding' in coptions._asdict():
            opts.update({'bin_seeding': bool(coptions.bin_seeding)})
        if 'min_bin_freq' in coptions._asdict():
            opts.update({'min_bin_freq': int(coptions.min_bin_freq)})

    def birch_opts():
        if 'threshold' in coptions._asdict():
            opts.update({'threshold': float(coptions.threshold)})
        if 'branching_factor' in coptions._asdict():
            opts.update({'branching_factor': int(coptions.branching_factor)})
        if 'n_clusters' in coptions._asdict():
            opts.update({'n_clusters': int(coptions.n_clusters)})

    def barber_opts():
        abo = config.get_stage_options(2)
        opts.update(get_auto_barber_options(abo))
        opts.update(get_auto_barber_options(coptions))

    if coptions.method == 'dbscan':
        dbscan_opts()
        method = DBSCAN
    elif coptions.method == 'affprop':
        affprop_opts()
        method = AffinityPropagation
    elif coptions.method == 'kmeans':
        kmeans_opts()
        method = KMeans
    elif coptions.method == 'meanshift':
        meanshift_opts()
        method = MeanShift
    elif coptions.method == 'birch':
        birch_opts()
        method = Birch
    elif coptions.method == 'barber':
        barber_opts()
        method = BarberCluster

    return PerformClustering(method, **opts)


def get_linearize_method(loption):
    if loption:
        assert isinstance(loption, str), "Wrong Linearize method definition: %r" % loption
        possible_formats = [
            re.compile('^(recursive|oneway|hobbit)(triangle|vector)[(][+-]?(\d+(\.\d*)?|\.\d+)([eE][+-]?\d+)?[)]$'),
            re.compile('^(recursive|oneway|hobbit)(triangle|vector)[(][)]$'),
            re.compile('^(recursive|oneway|hobbit)(triangle|vector)$')]
        assert True in [pf.match(loption.lower()) is not None for pf in
                        possible_formats], "Wrong Linearize method definition: %s" % loption
        # http://stackoverflow.com/questions/12929308/python-regular-expression-that-matches-floating-point-numbers#12929311
        way = [w for w in ['recursive', 'oneway', 'hobbit'] if w in loption.lower()][0]
        crit = [c for c in ['triangle', 'vector'] if c in loption.lower()][0]
        threshold = re.compile('[+-]?(\d+(\.\d*)?|\.\d+)([eE][+-]?\d+)?').findall(loption)
        if len(threshold):
            threshold = float(threshold[0][0])
        else:
            threshold = None
        # get method
        if way == 'recursive':
            if crit == 'triangle':
                met = traces.LinearizeRecursiveTriangle
            else:
                met = traces.LinearizeRecursiveVector
        elif way == 'oneway':
            if crit == 'triangle':
                met = traces.LinearizeOneWayTriangle
            else:
                met = traces.LinearizeOneWayVector
        elif way == 'hobbit':
            if crit == 'triangle':
                met = traces.LinearizeHobbitTriangle
            else:
                met = traces.LinearizeHobbitVector
        if threshold is None:
            return met()
        return met(threshold)


def discard_short_etc(spaths, short_paths=None, short_object=None, short_logic=None):
    # worker for discarding paths, can be time consuming because of filling coords cache
    if short_object is not None:
        return [sp for sp in spaths if
                short_logic(sp.size > short_paths, sp.object_len > short_object)] + [CRIC]
        # return len(spaths), [sp for sp in spaths if
        #                     short_logic(sp.size > short_paths, sp.object_len > short_object)], CRIC
    else:
        return [sp for sp in spaths if sp.size > short_paths]
        # return len(spaths), [sp for sp in spaths if sp.size > short_paths]


def center_of_object(spath):
    # import ipdb as pdb; pdb.set_trace()
    return 1, spath.center_of_object, CRIC


def get_allow_size_function(rt=None):
    if not isinstance(rt, str):
        return None
    assert re.compile('^[<>=]+[0-9.]+$').match(rt) is not None, "Wrong threshold definition: %s" % rt
    op = re.compile('[<>=]+')
    op = ''.join(sorted(op.findall(rt)[0]))
    vl = re.compile('[0-9.]+')
    vl = float(vl.findall(rt)[0])
    operator_dict = {'>': operator.gt,
                     '=>': operator.ge,
                     '<=': operator.le,
                     '<': operator.lt}
    assert op in list(operator_dict.keys()), "Unsupported operator %s in threshold %s" % (op, rt)
    return lambda size_of_cluster: operator_dict[op](size_of_cluster, vl)


def potentially_recursive_clustering(config=None,
                                     clustering_name=None,
                                     inlets_object=None,
                                     spaths=None,
                                     message='clustering',
                                     deep=0,
                                     max_level=5):
    with clui.fbm("Performing %s, level %d of %d" % (message, deep, max_level), cont=False):
        logger.debug('Clustering options section: %s' % clustering_name)
        cluster_options = config.get_cluster_options(section_name=clustering_name)
        clui.message('Clustering options:')
        for k, v in cluster_options._asdict().items():
            clui.message("%s = %s" % (str(k), str(v)))
        # TODO: Print clustering options in a nice way!
        clustering_function = get_clustering_method(cluster_options, config)
        # special case of barber!!!
        if cluster_options.method == 'barber':
            logger.debug('Getting inltets refs...')
            inlets_refs = set(inlets_object.get_inlets_references())
            logger.debug('Starting wtc...')
            wtc = WhereToCut(spaths=(sp for sp in spaths if sp.id in inlets_refs),
                             expected_nr_of_spaths=len(inlets_refs),
                             forceempty=True,
                             **clustering_function.method_kwargs)
            # clouds = wtc.cloud_groups(progress=True)
            logger.debug('Getting spheres...')
            inlets_object.add_spheres(wtc.spheres)
        logger.debug('Proceed with clustering, skip size...')
        # get skip_size function according to recursive_treshold
        skip_size = SkipSizeFunction(cluster_options.recursive_threshold)
        logger.debug('Proceed with clustering, call method...')
        inlets_object.perform_reclustering(clustering_function, skip_outliers=True, skip_size=skip_size)
    clui.message('Number of clusters detected so far: %d' % len(inlets_object.clusters_list))

    if cluster_options.recursive_clustering:
        deep += 1
        if deep > max_level:
            return
        return potentially_recursive_clustering(config=config,
                                                clustering_name=cluster_options.recursive_clustering,
                                                inlets_object=inlets_object,
                                                spaths=spaths,
                                                deep=deep,
                                                max_level=max_level)


def make_line(template, line):
    if len(template) != len(line):
        pass
    # detect nan and int problem
    for nr, (t, l) in enumerate(zip(template, line)):
        if is_number(l):
            if np.isnan(l):
                if 'd' in t:
                    template[nr] = t.replace('d', 's')
                    line[nr] = 'nan'
    return (' '.join(template)) % tuple(line)


def make_header_template(line_template):
    header_template = []
    col_re = re.compile('[0-9]+')
    for l in line_template:
        header_template.append('%{0}s'.format(col_re.findall(l)[0]))
    return header_template


def nr_header():
    return ['Nr'], ['%7d']


def get_header_line_and_line_template(header_line_and_line_template, head_nr=False):
    header, line_template = header_line_and_line_template

    header_template = make_header_template(line_template)
    # head_nr? only to header
    if head_nr:
        nrh, nrlt = nr_header()
        header_template = make_header_template(nrlt + line_template)
        header = nrh + header
    header_line = make_line(header_template, header)

    return header_line, line_template


def is_pymol_connector_session(filename):
    session_ext = re.compile('[.][pP][sS][eE]')
    if filename:
        ext = os.path.splitext(filename)[-1]
        if session_ext.match(ext):
            return True
    return False


def is_pymol_connector_script(filename):
    script = re.compile('.*[.]([pP][yY]|[.pP][yY][.][gG][zZ])$')
    if filename:
        if script.match(filename):
            return True
    return False


class SkipSizeFunction(object):

    def __init__(self, ths_def):

        self.thresholds = []
        if isinstance(ths_def, str):
            for thd in ths_def.split():
                self.thresholds.append(get_allow_size_function(thd))

    def __call__(self, size_of_cluster):
        for thd in self.thresholds:
            if not thd(size_of_cluster):
                return True
        return False


class PrintAnalysis(object):
    nr_template = '%7d '

    # TODO: Change it in such a way that it cooperates well with debug-file option.
    def __init__(self, fileoption, line_nr=False):
        self.output2stderr = False
        if fileoption:
            self.filehandle = open(fileoption, 'w')
            # self.output2stderr = True
        else:
            self.filehandle = sys.stdout
        self.line_nr = line_nr

    def __call__(self, info2print, nr=None):
        if self.line_nr and nr is not None:
            info2print = (self.nr_template % (nr + 1)) + info2print
        if self.output2stderr:
            clui.message(info2print)
        print(info2print, file=self.filehandle)

    def sep(self):
        self(asep())

    def thead(self, info2print):
        self(clui.thead(info2print))

    def tend(self, info2print):
        self(clui.tsep(info2print))

    def under(self, info2print):
        self(clui.underline(info2print))


def results_n(rn):
    if isinstance(rn, np.ndarray):
        return rn
    else:
        return np.memmap(rn[0], mode='r', dtype=np.int8, shape=rn[1])



# -*- coding: utf-8 -*-

# Aqua-Duct, a tool facilitating analysis of the flow of solvent molecules in molecular dynamic simulations
# Copyright (C) 2016-2018  Tomasz Magdziarz, Alicja Płuciennik, Michał Stolarczyk <info@aquaduct.pl>
# Copyright (C) 2019  Tomasz Magdziarz <info@aquaduct.pl>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np

import matplotlib.pyplot as plt

from aquaduct.geom import traces
from aquaduct.utils.helpers import list_blocks_to_slices
from aquaduct.visual.helpers import color_codes, cc


# matplotlib specific
# mpl color converter
# cc = lambda c,alpha=1.0 : colorConverter.to_rgba(c,alpha=alpha)


def yield_spath_len_and_smooth_diff_in_types_slices(sp, smooth=None, smooth_len=None, smooth_diff=None, types='etypes'):
    if smooth is not None:
        smooth_len = smooth
        smooth_diff = smooth
    # len
    coords = sp.get_coords_cont(smooth=smooth_len)
    dif = traces.diff(coords)
    ldif = np.cumsum(dif)
    # diff
    if smooth_diff != smooth_len:
        coords = sp.get_coords_cont(smooth=smooth_diff)
        dif = traces.diff(coords)
    # ldif = np.array(range(len(dif))) # same OX

    if types == 'etypes':
        etypes = sp.etypes_cont
    elif types == 'types':
        etypes = sp.types_cont

    for sl in list_blocks_to_slices(etypes):
        etype = etypes[sl]
        ld = ldif[sl]
        sd = dif[sl]
        while len(etype) > len(ld):
            etype.pop(-1)
        yield ld, sd, etype


def plot_colorful_lines(x, y, c, **kwargs):
    sls = list(list_blocks_to_slices(c))
    n = len(sls)

    for nr, sl in enumerate(sls):
        a = x[sl]
        b = y[sl]
        color = c[sl][-1]
        if nr == 0:
            plt.plot(a, b, color=color, **kwargs)
            last_a = a[-1]
            last_b = b[-1]
            last_color = color
        else:
            mid_a = (last_a + a[0]) / 2
            mid_b = (last_b + b[0]) / 2
            if nr == n - 1:
                plt.plot([last_a, mid_a], [last_b, mid_b], color=last_color, **kwargs)
                plt.plot([mid_a, a[0]], [mid_b, b[0]], color=color, **kwargs)
                plt.plot(a, b, color=color, **kwargs)
            else:
                plt.plot([last_a, mid_a], [last_b, mid_b], color=last_color, **kwargs)
                plt.plot([mid_a, a[0]], [mid_b, b[0]], color=color, **kwargs)
                plt.plot(a, b, color=color, **kwargs)
                last_a = a[-1]
                last_b = b[-1]
                last_color = color


def spaths_spectra(spaths, **kwargs):
    spectra = []

    minx = None
    maxx = None
    for sp in spaths:
        xyc = list(spath_spectrum(sp, **kwargs))
        xy = np.array([(x, y) for x, y, c in xyc])
        c = list([c for x, y, c in xyc])
        x = xy[:, 0]
        y = xy[:, 1]
        spectra.append((y, c))
        if maxx is None:
            maxx = max(x)
        else:
            maxx = max(maxx, max(x))
        if minx is None:
            minx = min(x)
        else:
            minx = min(minx, min(x))
    for (y, c) in spectra:
        plot_colorful_lines(np.linspace(minx, maxx, len(c)), y, c)


def plot_spath_spectrum(sp, **kwargs):
    xyc = list(spath_spectrum(sp, **kwargs))
    xy = np.array([(x, y) for x, y, c in xyc])
    c = list([c for x, y, c in xyc])
    plot_colorful_lines(xy[:, 0], xy[:, 1], c)


def spath_spectrum(sp, **kwargs):
    lsdt = list(yield_spath_len_and_smooth_diff_in_types_slices(sp, **kwargs))

    n = len(lsdt)

    last_l = None
    last_sd = None
    last_t = None
    last_color = None
    for nr, (l, sd, t) in enumerate(lsdt):
        color = color_codes(t[-1])
        if nr == 0:
            for ll, ssdd in zip(l, sd):
                yield ll, ssdd, color
            # plt.plot(l, sd, color=color)
            last_l = l[-1]
            last_sd = sd[-1]
            last_color = color
        else:
            mid_l = (last_l + l[0]) / 2
            mid_sd = (last_sd + sd[0]) / 2
            if nr == n - 1:
                # for ll, ssdd in zip([last_l, mid_l], [last_sd, mid_sd]):
                #    yield ll, ssdd, last_color
                # plt.plot([last_l, mid_l], [last_sd, mid_sd], color=last_color)
                # for ll, ssdd in zip([mid_l, l[0]], [mid_sd, sd[0]]):
                #    yield ll, ssdd, color
                # plt.plot([mid_l, l[0]], [mid_sd, sd[0]], color=color)
                for ll, ssdd in zip(l, sd):
                    yield ll, ssdd, color
                    # plt.plot(l, sd, color=color)
            else:
                # for ll, ssdd in zip([last_l, mid_l], [last_sd, mid_sd]):
                #    yield ll, ssdd, last_color
                # plt.plot([last_l, mid_l], [last_sd, mid_sd], color=last_color)
                # for ll, ssdd in zip([mid_l, l[0]], [mid_sd, sd[0]]):
                #    yield ll, ssdd, color
                # plt.plot([mid_l, l[0]], [mid_sd, sd[0]], color=color)
                for ll, ssdd in zip(l, sd):
                    yield ll, ssdd, color
                # plt.plot(l, sd, color=color)
                last_l = l[-1]
                last_sd = sd[-1]
                last_color = color


def showit(gen):
    def patched(*args, **kwargs):
        obj = gen(*args, **kwargs)
        plt.show()
        return obj

    return patched


def get_ax3d(fig, sub=111):
    return fig.add_subplot(sub, projection='3d')


class SimpleTracePlotter(object):
    def plot_line(self, coords, color, **kwargs):
        raise NotImplementedError('This is abstract class. Missing implementaion in a child class.')

    '''
    def path_trace(self,path,color=('r','g','b'),
                   plot_in=True,
                   plot_object=True,
                   plot_out=True,
                   **kwargs):
        raise NotImplementedError('This is base class.')
    '''

    def single_trace(self, coords, color='r', **kwargs):
        # coords is a trace
        # color is a single color
        color = cc(color)
        coords = np.array(coords)
        # call plot_line
        self.plot_line(coords, color, **kwargs)

    def path_trace(self, path, color=('r', 'g', 'b'),
                   plot_in=True,
                   plot_object=True,
                   plot_out=True,
                   **kwargs):
        # path is a tuple of length 3, its elements represent in,object, out parts of path
        # color is a tuple of length 3, its elements correspond to colors of consecutive path parts
        color = list(map(cc, color))
        for nr, trace in enumerate(traces.midpoints(path)):
            # mid points!
            if len(trace) > 0:
                if (nr == 0 and plot_in) or (nr == 1 and plot_object) or (nr == 2 and plot_out):
                    self.single_trace(trace, color=color[nr], **kwargs)


class SimpleProteinPlotter(SimpleTracePlotter):
    def protein_trace(self, protein, smooth=None, color=('c', 'm', 'y'), **kwargs):
        # assumes protein is reader object
        # TODO: iterate over chains?
        bb = protein.parse_selection("protein and backbone")
        coords = bb.get_positions()
        cdiff = traces.diff(coords)
        split = np.argwhere(cdiff > 2.5)  # TODO: magic constant
        ns = len(split)
        if ns == 0:
            self.single_trace(smooth(coords), color=color[0], **kwargs)  # TODO: color conversion is buggy
        else:
            split.shape = (ns,)
            for nr, csplit in enumerate([0] + split.tolist()):
                cc = color[nr % len(color)]
                if nr == ns:
                    scoords = coords[csplit + 1:]
                elif nr == 0:
                    scoords = coords[csplit:split[nr]]
                else:
                    scoords = coords[csplit + 1:split[nr]]
                if smooth:
                    scoords = smooth(scoords)
                self.single_trace(scoords, color=color, **kwargs)


class SimplePathPlotter(SimpleTracePlotter):
    def single_path_traces(self, spaths, smooth=None, color=('r', 'g', 'b'), **kwargs):
        for spath in spaths:
            self.path_trace(spath.get_coords(smooth), color=color, **kwargs)


class MPLTracePlotter(SimplePathPlotter, SimpleProteinPlotter):
    @showit
    def init_ax(self):
        self.fig = plt.figure()
        self.ax = self.fig.add_subplot(111, projection='3d')

        self.fig.subplots_adjust(left=0, bottom=0, right=1, top=1)
        self.fig.set_facecolor('w')

        self.ax.set_axis_bgcolor('none')
        self.ax.axis('off')

    @showit
    def plot_line(self, coords, color, **kwargs):
        self.ax.plot3D(coords[:, 0],
                       coords[:, 1],
                       coords[:, 2],
                       c=color, **kwargs)

    @showit
    def scatter(self, coords, **kwargs):
        self.ax.scatter3D(coords[:, 0],
                          coords[:, 1],
                          coords[:, 2],
                          **kwargs)



# -*- coding: utf-8 -*-

# Aqua-Duct, a tool facilitating analysis of the flow of solvent molecules in molecular dynamic simulations
# Copyright (C) 2018-2019  Michał Banas <info@aquaduct.pl>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Module used for caching results of specific function using database to store data"""

from hashlib import md5
from aquaduct import logger
import numpy as np


class DBCache(object):
    def __init__(self, db_dir):
        self.db_dir = db_dir

    def __call__(self, func, *args, **kwargs):
        def wrapper(*args, **kwargs):
            key = md5(','.join(map(str, args)) + '&' + ','.join(
                [':'.join(map(str, kv)) for kv in iter(kwargs.items())])).hexdigest()

            logger.debug('Looking for cache key {}'.format(key))

            try:
                coords = np.load(self.db_dir + key + ".npy")
                print("Data fetched")
                return coords
            except IOError:
                coords = func(*args, **kwargs)

                np.save(self.db_dir + key, coords)

                logger.debug("New key added to cache.")

                return coords

        return wrapper



#!/usr/bin/env python3.7
# -*- coding: utf-8 -*-

# Aqua-Duct, a tool facilitating analysis of the flow of solvent molecules in molecular dynamic simulations
# Copyright (C) 2018-2019  Michał Banas
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import tkinter as tk
import hashlib
import traceback
import os
import shutil
import tempfile
import tkinter.messagebox
import tkinter.ttk
import webbrowser
from configparser import ConfigParser, NoOptionError, NoSectionError
from collections import OrderedDict, defaultdict
from tkinter.filedialog import askopenfile, askdirectory, asksaveasfile

import aquaduct
import aquaduct.apps.valveconfig.defaults as defaults
import aquaduct.apps.valveconfig.utils as utils
from aquaduct.apps.valveconfig import get_img


class ValveConfigApp(object):
    def __init__(self, parent):
        """ Valve Configurator App """
        self.size = (600, 650)

        self.parent = parent

        self.values = OrderedDict()

        # Used to hide that frame, after all entries are shown
        self.init_frame = tkinter.ttk.Frame(self.parent)
        self.notebook = tkinter.ttk.Notebook(self.parent)

        self.level = None

        self.hiding_frames = {}

        self.cluster_frame_index = None

        self.recluster_frame_index = None

        # Used for identifying scroll frame by interior frame for removing it from notebook
        self.scrolled_frames = {}

        self.frames = []

        # Dict with input frames of required entries. Used to change color if is unfilled
        self.required_entries = defaultdict(dict)

        self.config_filename = tk.StringVar()
        # Used to determine if max_level difference message should be displayed
        self.config_loaded = False

        self.values_hash = self.get_values_hash()

        self.title = "Valve Configurator"

        parent.title(self.title)
        parent.geometry("600x650")

        self.parent.protocol("WM_DELETE_WINDOW", self.on_window_close)

        self.init_gui()

    def init_gui(self):
        """ Prepare initial frame """
        # Logo
        logo = tk.PhotoImage(file=get_img("logo.gif"))

        logo_label = tkinter.ttk.Label(self.parent, image=logo, padding=-2)
        logo_label.image = logo
        logo_label.pack(padx=20, pady=20)

        # Used to auto positioning depending on length of version string
        version = "ver. " + aquaduct.version_nice()
        tkinter.ttk.Label(logo_label, text=version, background="white").place(relx=1, rely=1, x=-len(version) * 7, y=-20)

        # Frame with loading configuration file
        load_frame = tkinter.ttk.LabelFrame(self.init_frame, text="Configuration file")
        load_frame.columnconfigure(0, weight=1)
        load_frame.columnconfigure(1, weight=1)

        tkinter.ttk.Button(load_frame, text="Load config", command=self.open_config_file).grid(sticky="E", row=0, column=0)
        tkinter.ttk.Button(load_frame, text="New config", command=self.create_new_config_file_dialog).grid(sticky="W",
                                                                                                   row=0,
                                                                                                   column=1,
                                                                                                   padx=5,
                                                                                                   pady=2)
        tkinter.ttk.Entry(load_frame, textvariable=self.config_filename, state="readonly").grid(sticky="EW",
                                                                                        row=1,
                                                                                        column=0,
                                                                                        columnspan=2,
                                                                                        padx=5,
                                                                                        pady=2)

        # Level selection
        level_frame = tkinter.ttk.LabelFrame(self.init_frame, text="Configuration level")

        self.level = tk.StringVar()

        # List of sorted levels from easiest to hardest
        self.levels = []
        for level_name, _ in reversed(sorted(iter(defaults.LEVELS.items()), key=lambda k_v: (k_v[1], k_v[0]))):
            self.levels.append(level_name)

        tkinter.ttk.OptionMenu(level_frame, self.level, self.levels[0], *self.levels).pack(pady=20)

        load_frame.pack(fill=tk.X, padx=100, pady=20)
        level_frame.pack(fill=tk.X, padx=100, pady=20)

        def show_new_gui():
            self.prepare_section_frames()
            self.create_interface()

        tkinter.ttk.Button(self.init_frame, text="Forward", command=show_new_gui).pack(pady=50)

        self.init_frame.pack(expand=1, fill="both")

    def prepare_section_frames(self):
        """ Parse DEFAULTS and depends on it create Notebook tabs and entries """
        for section in defaults.DEFAULTS:
            if section._nested: # Skip nested frames which were added in runtime.
                continue

            section_name = section.config_name
            section_name_long = section.name
            entries = section.entries

            if section.abs_level is not None:
                if section.abs_level != defaults.LEVELS[self.level.get()]:
                    continue

            if defaults.LEVELS[self.level.get()] > section.level:
                continue

            if section_name == "clustering":
                self.cluster_frame_index = len(self.frames)

            if section_name == "reclustering":
                self.recluster_frame_index = len(self.frames)

            scroll_frame = utils.VerticalScrolledFrame(self.notebook)

            frame = scroll_frame.interior
            frame.columnconfigure(0, weight=1)
            frame.columnconfigure(1, weight=1)

            self.frames.append(frame)
            self.scrolled_frames[frame] = scroll_frame

            self.notebook.add(scroll_frame, text=section_name_long)

            tkinter.ttk.Label(frame,
                      text=section_name_long,
                      style="Title.TLabel",
                      background=utils.get_widget_bg(self.parent)).grid(row=0, column=0, columnspan=2)
            tkinter.ttk.Separator(frame, orient=tk.HORIZONTAL).grid(sticky="EW", row=1, column=0, columnspan=3)

            self.entry_filler(frame, section_name, entries)

        if self.cluster_frame_index:
            for section_name in self.get_recursive_clustering_sections("clustering"):
                self.append_entries(section_name)

                # Increment max_level
                self.add_max_level(1)

            cluster_add_button = tkinter.ttk.Button(self.frames[self.cluster_frame_index], text="Add clustering section")
            # Setting row=1000 let skip calculating position of button each time new section is added
            cluster_add_button.grid(row=1000, column=0, columnspan=2, pady=20)

            cluster_add_section_callback = utils.CallbackWrapper(self.callback_add_section, "clustering")
            cluster_add_button.bind("<Button-1>", cluster_add_section_callback)

        # if "reclustering" in self.hiding_frames:
        if self.recluster_frame_index:
            for section_name in self.get_recursive_clustering_sections("reclustering"):
                self.append_entries(section_name)

            recluster_add_button = tkinter.ttk.Button(self.frames[self.recluster_frame_index], text="Add reclustering section")
            # Setting row=1000 let skip calculating position of button each time new section is added
            recluster_add_button.grid(row=1000, column=0, columnspan=2, pady=20)

            recluster_add_section_callback = utils.CallbackWrapper(self.callback_add_section, "reclustering")
            recluster_add_button.bind("<Button-1>", recluster_add_section_callback)

        # Refresh all hiding frames
        self.refresh_menus()

        self.init_frame.pack_forget()

        self.notebook.pack(expand=1, fill="both")

        # After preparing all section load values from config
        if self.config_filename.get() != "":
            self.load_config_values(self.config_filename.get())

        self.values_hash = self.get_values_hash()

    def create_interface(self):
        """ Creates window menu """
        menu_bar = tk.Menu(self.parent)

        # Level menu
        level_menu = tk.Menu(menu_bar)

        def change_level_callback(level_name):
            self.level.set(level_name)
            self.recreate_gui()

        for level_name in self.levels:
            level_cb = utils.CallbackWrapper(change_level_callback, level_name)
            level_menu.add_command(label=level_name, command=level_cb)

        # File menu
        file_menu = tk.Menu(menu_bar, tearoff=0)
        file_menu.add_command(label="New", command=self.create_new_config_file_dialog)

        def open_config():
            if self.open_config_file():
                self.load_config_values(self.config_filename.get())

                self.values_hash = self.get_values_hash()

        file_menu.add_command(label="Open", command=open_config)

        def save_callback(*args):
            if self.config_filename.get() == "":
                f = asksaveasfile("w")
                f.close()
                if f and self.save_config(f.name):
                    self.parent.title("{} - {}".format(self.title, f.name))
                    self.config_filename.set(f.name)
                else:
                    os.remove(f.name)
            else:
                self.save_config(self.config_filename.get())

        file_menu.add_command(label="Save", command=save_callback)

        def save_as_callback(*args):
            f = asksaveasfile("w")
            if f and self.save_config(f.name) and self.config_filename.get() == "":
                self.parent.title("{} - {}".format(self.title, f.name))
                self.config_filename.set(f.name)

        file_menu.add_command(label="Save as", command=save_as_callback)

        file_menu.add_separator()

        reset_cb = utils.CallbackWrapper(self.load_defaults, reset=True)
        file_menu.add_command(label="Reset to default", command=reset_cb)
        file_menu.add_cascade(label="Level", menu=level_menu)
        file_menu.add_command(label="Exit", command=self.on_window_close)
        menu_bar.add_cascade(label="File", menu=file_menu)

        # Run menu
        run_menu = tk.Menu(menu_bar, tearoff=0)
        run_menu.add_command(label="Valve", command=self.valve_run_dialog)
        # run_menu.add_command(label="Pond", command=self.pond_run_dialog)
        menu_bar.add_cascade(label="Run", menu=run_menu)

        # Help menu
        help_menu = tk.Menu(menu_bar, tearoff=0)
        help_menu.add_command(label="Documentation", command=self.open_docs)
        help_menu.add_command(label="About", command=self.about)
        menu_bar.add_cascade(label="Help", menu=help_menu)

        self.parent.config(menu=menu_bar)

    def callback_add_section(self, section_name):
        """
        Callback for button to create new recursive clustering in clustering/reclustering frame

        :param section_name: Config section name where new frames will be appended, allowed are "clustering" or "reclustering"
        """
        # Find free name for new section
        index = 0
        while section_name + str(index) in self.values:
            index += 1

        self.append_entries(section_name + str(index))

        if section_name.startswith("clustering"):
            self.add_max_level(1)

        self.refresh_menus()

    def callback_remove_section(self, section_name, frame):
        """
        Callback for button to remove existing recursive clustering in clustering/reclustering frame

        :param section_name: Config section name to remove, allowed are "clustering" or "reclustering"
        :param frame: frame Which contains options related to that section
        """
        del self.values[section_name]
        del self.hiding_frames[section_name]

        # Remove from MENUS list
        for i, item in enumerate(defaults.MENUS):
            if item.startswith(section_name):
                del defaults.MENUS[i]

        if section_name.startswith("clustering"):
            self.add_max_level(-1)

        frame.grid_forget()

    def get_recursive_clustering_sections(self, section_name):
        """
        Finds recursively all section used in recursive_clustering options

        :param section_name: Config section name from which fetching will start.
        :return Uniqe section names without section_name param.
        :rtype: list
        """
        if section_name != "clustering" and section_name != "reclustering":
            raise RuntimeError(
                "There is no possibility to get recursive clustering from {} section".format(section_name))

        try:
            with open(self.config_filename.get(), "r") as config_file:
                config = ConfigParser()
                config.readfp(config_file)

        except IOError:  # File is empty
            return []

        clustering_sections = []
        try:
            clustering_section = config.get(section_name, "recursive_clustering")
            while clustering_section not in clustering_sections and clustering_section != "None":
                clustering_sections.append(clustering_section)
                clustering_section = config.get(clustering_section, "recursive_clustering")
        except (NoOptionError, NoSectionError):
            pass

        return list(set(clustering_sections).difference([section_name]))

    def append_entries(self, section_name):
        """
        Append new frame with new clustering or reclustering options.

        :param section_name: Config section name where new frames will be appended, allowed are "clustering" or "reclustering".
        """
        if section_name.startswith("clustering") and not self.cluster_frame_index:
            return

        if section_name.startswith("reclustering") and not self.recluster_frame_index:
            return

        if section_name.startswith("clustering"):
            default_section_name = "clustering"
            frame = self.frames[self.cluster_frame_index]
        elif section_name.startswith("reclustering"):
            default_section_name = "reclustering"
            frame = self.frames[self.recluster_frame_index]
        else:
            raise RuntimeError(
                "Appending entries to sections other than clustering or reclusteriation is not allowed")

        # Calculate row
        row = len(frame.grid_slaves(None, 0)) + 2

        # Set option menu from recursive clustering to control hiding frames
        defaults.MENUS.append("{}:method".format(section_name))

        default_entries = defaults.get_default_section(default_section_name)

        inner_frame = tkinter.ttk.Frame(frame, style="I.TFrame")
        inner_frame.columnconfigure(0, weight=1)
        inner_frame.columnconfigure(1, weight=1)
        inner_frame.grid(sticky="EW", row=row, column=0, columnspan=2, padx=60, pady=10)

        tkinter.ttk.Label(inner_frame, text="Recursive clustering", style="ITitle.TLabel").grid(
            row=0,
            column=0,
            pady=5,
            columnspan=2)
        tkinter.ttk.Separator(inner_frame, orient=tk.HORIZONTAL).grid(sticky="EW", row=1, column=0, columnspan=2)

        self.entry_filler(inner_frame, section_name, default_entries.entries)

        remove_button = tkinter.ttk.Button(inner_frame, text="Remove")
        remove_button.grid(row=999, column=0, columnspan=2, pady=5)

        remove_section_callback = utils.CallbackWrapper(self.callback_remove_section, section_name, inner_frame)
        remove_button.bind("<Button-1>", remove_section_callback)

    def entry_filler(self, parent, section_name, entries):
        """
        Creates entries depending on DEFAULTS values.

        When no entry added it hide notebook tab.

        Disables name option for clustering, reclustering and derivatives section.

        :param parent: Parent widget.
        :param section_name: Config section name.
        :param entries: List of entries.

        If section is nested, it's level is not checked. It allow to nest section that have higher level,
        but it cannot be visible in notebook.
        """
        # Keeps label frames with actual row
        group_frames = {}

        _, row = parent.grid_size()

        entries_appended = 0
        for entry in entries:
            # If its primary reclustering or clustering section disable changing name of that section
            # Additionaly it disable inlets_clustering:max_level if config level is higher than max_level level
            state = tk.NORMAL
            if section_name == "clustering" and entry.config_name == "name":
                state = tk.DISABLED
            elif section_name == "reclustering" and entry.config_name == "name":
                state = tk.DISABLED
            elif section_name == "inlets_clustering" and entry.config_name == "max_level":
                if defaults.LEVELS[self.level.get()] != entry.level:
                    state = tk.DISABLED

            # Skip entries that dont match level except inlets_clustering:max_level
            # and are not nested section
            if defaults.LEVELS[self.level.get()] > entry.level and not isinstance(entry, defaults.DefaultSection):
                if not (section_name == "inlets_clustering" and entry.config_name == "max_level"):
                    continue

            entries_appended += 1

            if isinstance(entry, defaults.DefaultSection):
                nested_frame = tkinter.ttk.LabelFrame(parent, text=entry.name)
                nested_frame.grid_columnconfigure(0, weight=1)
                nested_frame.grid_columnconfigure(1, weight=1)

                """
                Adding in runtime section to DEFAULTS sections to let them be visible by 
                defaults.get_default_entry or defaults.get_default_section functions during 
                reloading values when level was changed.
                """
                entry._nested = True
                defaults.DEFAULTS.append(entry)
                self.entry_filler(nested_frame, entry.config_name, entry.entries)

                if nested_frame.grid_size()[1]:
                    nested_frame.grid(row=parent.grid_size()[1], column=0, columnspan=2, pady=15, ipadx=30)
            elif entry.optionmenu_value:
                if section_name not in self.hiding_frames:
                    self.hiding_frames[section_name] = {}

                if entry.optionmenu_value not in self.hiding_frames[section_name]:
                    self.hiding_frames[section_name][entry.optionmenu_value] = utils.HidingFrame(parent,
                                                                                                 parent.grid_size()[1],
                                                                                                 text=entry.optionmenu_value.capitalize() + " options",
                                                                                                 style="HF.TFrame")

                    # It must be showed first, to allocate space in frame, otherwise grid_size wont return proper value
                    self.hiding_frames[section_name][entry.optionmenu_value].show()

                hiding_frame = self.hiding_frames[section_name][entry.optionmenu_value]

                if section_name not in self.values:
                    self.values[section_name] = {}

                if entry.optionmenu_value not in self.values[section_name]:
                    self.values[section_name][entry.optionmenu_value] = {}

                self.values[section_name][entry.optionmenu_value][entry.config_name] = utils.entry_factory(hiding_frame,
                                                                                                           hiding_frame.grid_size()[1],
                                                                                                           entry.name,
                                                                                                           entry.default_values,
                                                                                                           entry.help_text,
                                                                                                           state,
                                                                                                           info_text=entry.info_text,
                                                                                                           warning_text=entry.warning_text)

                if entry.required:
                    self.required_entries[section_name][entry.config_name] = self.values[section_name][
                        entry.config_name]
            else:
                if entry.group_label:
                    if entry.group_label not in group_frames:
                        label_frame = tkinter.ttk.LabelFrame(parent, text=entry.group_label)
                        label_frame.grid_columnconfigure(0, weight=1)
                        label_frame.grid_columnconfigure(1, weight=1)
                        label_frame.grid(row=parent.grid_size()[1], column=0, columnspan=2, pady=15, ipadx=30)

                        group_frames[entry.group_label] = [label_frame, 0]

                    entry_parent = group_frames[entry.group_label][0]
                    entry_row = group_frames[entry.group_label][1]

                    group_frames[entry.group_label][1] += 1
                else:
                    entry_parent = parent

                entry_widget = utils.entry_factory(entry_parent,
                                                   entry_parent.grid_size()[1],
                                                   entry.name,
                                                   entry.default_values,
                                                   entry.help_text,
                                                   state,
                                                   info_text=entry.info_text,
                                                   warning_text=entry.warning_text)

                if (section_name.startswith("clustering") or
                    section_name.startswith("reclustering")) and entry.config_name == "name":
                    entry_widget.set(section_name)

                if defaults.is_menu(section_name, entry.config_name):
                    opt_menu_callback = utils.CallbackWrapper(self.option_menu_changed, section_name, entry_widget)
                    entry_widget.input_var.trace("w", opt_menu_callback)

                if section_name not in self.values:
                    self.values[section_name] = {}

                self.values[section_name][entry.config_name] = entry_widget

                if entry.required:
                    self.required_entries[section_name][entry.config_name] = entry_widget

        # Hide tab if unused
        if entries_appended == 0:
            if parent in self.scrolled_frames:  # Only for sections that are part of notebook
                self.notebook.forget(self.scrolled_frames[parent])

    def refresh_menus(self):
        """
        Sets visible only hiding frames that are chosen by option menu.
        """
        for menu in defaults.MENUS:
            section_name, entry_name = menu.split(":")

            if self.values.get(section_name, None):
                self.option_menu_changed(section_name, self.values[section_name][entry_name])

    def re_clustering_exists(self, section_name):
        """
        Returns true if clustering or reclustering sections exists

        It omit ability to create recursive sections

        :param section_name: Section name
        :type str
        :return: True if section exists, otherwise False
        :rtype bool
        """
        return section_name in self.hiding_frames

    def option_menu_changed(self, section, entry):
        """
        Callback for option menu that control hiding frames

        :param section: Section name where hiding frame is located.
        :param entry: Option menu widget, which control hiding frames in section.
        """
        for method, hidden_frame in self.hiding_frames[section].items():
            if method == entry.get():
                hidden_frame.show()
            else:
                hidden_frame.grid_forget()

    def add_max_level(self, value):
        """
        Add value to max_level

        :param value: Number which will be added to max_level.
        """
        current_value = self.values["inlets_clustering"]["max_level"].get()
        self.values["inlets_clustering"]["max_level"].set(current_value + value)

    def open_config_file(self):
        """ Show dialog to choose config file and save its name """
        filetypes = (("text files", "*.txt"), ("config files", "*.cfg"), ("all files", "*.*"))
        try:
            with askopenfile("r", filetypes=filetypes) as config_file:
                self.parent.title("{} - {}".format(self.title, config_file.name))
                self.config_filename.set(config_file.name)
        except AttributeError:
            return False

        return True

    def recreate_gui(self, *args):
        """
        Recreate GUI.

        Saves current values to temporary file.

        :param args: Event informations.
        """
        tmpfile = tempfile.NamedTemporaryFile(delete=False)
        tmpfile.close()
        tmp_filename = tmpfile.name

        # Save values to temporary file.
        self.save_config(tmpfile.name, required_checking=False)

        self.cluster_frame_index = None
        self.recluster_frame_index = None

        for tab_id in self.notebook.tabs():
            self.notebook.forget(tab_id)

        self.hiding_frames = {}

        self.prepare_section_frames()
        self.load_defaults()  # Reset values
        self.load_config_values(tmp_filename)

        os.remove(tmp_filename)

    def get_values_hash(self):
        """ Compute hash from values. """
        hash = hashlib.md5()
        hash.update((str(self.values)).encode('utf-8'))
        # for section_name in self.values:
        #     for value in self.values[section_name].itervalues():
        #         hash.update(str(value.get()))

        return hash.digest()

    def load_config_values(self, config_filename):
        """
        Load config values to values dictionary from file.

        Additionally it ask user if inlets_clustering:max_level in config
        is different from number of found recursive_clustering sections.

        :param config_filename: Name of file with configuration.
        """
        # Config can be loaded when UI was created.
        # Checking if there is no additional clustering or reclustering sections and add them.
        if self.re_clustering_exists("clustering"):
            for section_name in self.get_recursive_clustering_sections("clustering"):
                if not self.re_clustering_exists(section_name):
                    self.append_entries(section_name)

        with open(config_filename, "r") as config_file:
            config = ConfigParser()
            config.readfp(config_file)

            for section_name in self.values:
                for entry_name in self.values[section_name]:

                    # Additional for is required to handle clustering/reclustering methods
                    entries = [entry_name] if not isinstance(self.values[section_name][entry_name], dict) else self.values[section_name][entry_name]
                    for entry_name in entries:
                        try:
                            config_value = config.get(section_name, entry_name)

                            #  Inform user that config max_level differ from number of found recursive clustering
                            result = True
                            if section_name == "inlets_clustering" and entry_name == "max_level" and self.config_loaded:
                                if self.values["inlets_clustering"]["max_level"].get() != int(config_value):
                                    result = tkinter.messagebox.askyesno("Max level",
                                                                   "Config max_level value differs from number of found clustering sections.\n"
                                                                   "Do you want to keep value from configuration file?")

                            if not result:
                                continue

                            optionmenu_value = defaults.get_default_entry(section_name, entry_name).optionmenu_value
                            if optionmenu_value:
                                try:
                                    method = config.get(section_name, "method")
                                except NoOptionError:
                                    method = defaults.get_default_entry(section_name, "method")

                                self.values[section_name][method][entry_name].set(config_value)
                            else:
                                self.values[section_name][entry_name].set(config_value)
                        except NoOptionError:
                            continue
                        except NoSectionError:
                            break

        self.config_loaded = True

    def load_defaults(self, reset=False, *args):
        """
        Load defaults values to values dictionary.

        :param reset: If set to True confirmation will be needed to reset.
        """
        if reset:
            result = tkinter.messagebox.askyesno("Reset values", "Are you sure to reset all values?")

            if not result:
                return

        for section in defaults.DEFAULTS:
            if defaults.LEVELS[self.level.get()] > section.level:
                continue

            for entry_section, entry in section.iter_entries():
                # Skip entries that dont match level except inlets_clustering:max_level
                if defaults.LEVELS[self.level.get()] > entry.level:
                    if not (entry_section.config_name == "inlets_clustering" and entry.config_name == "max_level"):
                        continue

                if entry.optionmenu_value:
                    self.values[entry_section.config_name][entry.optionmenu_value][entry.config_name].set(entry.default_value)
                else:
                    self.values[entry_section.config_name][entry.config_name].set(entry.default_value)

        # Delete appended clustering and reclustering sections
        if self.cluster_frame_index:
            cluster_frame = self.frames[self.cluster_frame_index]
            for widget in cluster_frame.grid_slaves():
                if isinstance(widget, tkinter.ttk.Frame):
                    widget.grid_forget()

        if self.recluster_frame_index:
            recluster_frame = self.frames[self.recluster_frame_index]
            for widget in recluster_frame.grid_slaves():
                if isinstance(widget, tkinter.ttk.Frame):
                    widget.grid_forget()

        self.refresh_menus()

        # Delete rest informations about appended sections
        for section_name in self.values.keys():
            if section_name.startswith("clustering") or section_name.startswith("clustering"):
                if section_name != "clustering" and section_name != "reclustering":
                    del self.values[section_name]
                    del self.hiding_frames[section_name]

                    # Remove from MENUS list
                    for i, item in enumerate(defaults.MENUS):
                        if item.startswith(section_name):
                            del defaults.MENUS[i]

    def get_active_frame_name(self, section_name):
        """
        Return currently gridded HidingFrame in section_name.

        :param section_name: Name of section from which active HidingFrame will be returned.
        :return: Name of chosen option menu value.
        """
        for method, frame in self.hiding_frames[section_name].items():
            if any(frame.grid_info()):
                return method

    def save_config(self, config_filename, required_checking=True):
        """
        Save all values to config file.

        :param config_filename: Name of file where options will be saved.
        :param required_checking: If False checking required fields and saving dialog is skipped. Needed to save options before changing level.
        """
        # Unhighlight all frames
        for section_name in self.required_entries:
            for entry_name in self.required_entries[section_name]:
                self.required_entries[section_name][entry_name].unhighlight()

        if required_checking:
            # Check if all field all filled before saving
            tabs_id = self.notebook.tabs()
            first_found = [None, None, None]  # Keeps info of first found unfilled entry
            for i, section_name in enumerate(self.values):
                for entry_name in self.values[section_name]:
                    if isinstance(self.values[section_name][entry_name], dict):
                        for entry_name_optionmenu, entry_value_optionmenu in self.values[section_name][entry_name].items():
                            if not entry_value_optionmenu.get() and entry_name_optionmenu in self.required_entries[section_name]:
                                section_full_name = defaults.get_default_section(section_name).name
                                entry_full_name = defaults.get_default_entry(section_name, entry_value_optionmenu).name[:-2]

                                if not first_found[0]:
                                    first_found[0] = section_full_name
                                    first_found[1] = entry_full_name
                                    first_found[2] = tabs_id[i]

                                self.required_entries[section_name][entry_name].highlight()
                    else:
                        if not self.values[section_name][entry_name].get() and entry_name in self.required_entries[section_name]:
                            section_full_name = defaults.get_default_section(section_name).name
                            entry_full_name = defaults.get_default_entry(section_name, entry_name).name[:-2]

                            if not first_found[0]:
                                first_found[0] = section_full_name
                                first_found[1] = entry_full_name
                                first_found[2] = tabs_id[i]

                            self.required_entries[section_name][entry_name].highlight()

            if first_found[2]:  # 0 and 1 indexes is not checked because it can be just an empty string
                tkinter.messagebox.showerror("Unfilled field",
                                       "Field \"{}\" in \"{}\" must be specified.".format(entry_full_name,
                                                                                          section_full_name))
                self.notebook.select(first_found[2])
                return False

        # Create config file backup
        shutil.copy(config_filename, config_filename + ".bak")
        try:
            with open(config_filename, "w+") as config_file:
                config = ConfigParser()

                for section in defaults.DEFAULTS:
                    if section.additional:
                        continue

                    if section.config_name == "clustering" and not self.re_clustering_exists(section.config_name):
                        continue

                    if section.config_name == "reclustering" and not self.re_clustering_exists(section.config_name):
                        continue

                    if not config.has_section(section.config_name):
                        config.add_section(section.config_name)

                    for entry_section, entry in section.iter_entries():
                        value = None

                        # Try to get value from entry
                        try:
                            if entry.optionmenu_value:
                                value = self.values[entry_section.config_name][entry.optionmenu_value][entry.config_name].get()
                            else:
                                value = self.values[entry_section.config_name][entry.config_name].get()
                        except KeyError:
                            value = entry.default_value

                        if value == "" or (not isinstance(value, bool) and value == 0):
                            continue

                        if entry.optionmenu_value:
                            # Save options only from chosen options in optionmenu
                            if entry.optionmenu_value == self.get_active_frame_name(section.config_name):
                                config.set(section.config_name, entry.config_name, str(value))
                        else:
                            config.set(section.config_name, entry.config_name, str(value))

                # Add created clustering and reclustering sections
                clustering_section_names = [key for key in list(self.values.keys()) if
                                            key.startswith("clustering") and not key == "clustering"]
                reclustering_section_names = [key for key in list(self.values.keys()) if
                                              key.startswith("reclustering") and not key == "reclustering"]

                for section_name in clustering_section_names + reclustering_section_names:
                    section_config_name = self.values[section_name]["name"].get()
                    config.add_section(section_config_name)

                    for entry_section, entry in defaults.get_default_section(section_name).iter_entries():
                        value = None

                        # Try to get value from entry
                        try:
                            if entry.optionmenu_value:
                                value = self.values[section_name][entry.optionmenu_value][entry.config_name].get()
                            else:
                                value = self.values[section_name][entry.config_name].get()
                        except KeyError:
                            value = entry.default_value

                        if value == "" or value == 0:
                            continue

                        if entry.optionmenu_value:
                            # Save options only from chosen options in optionmenu
                            if entry.optionmenu_value == self.get_active_frame_name(section_name):
                                config.set(section_config_name, entry.config_name, str(value))
                        else:
                            config.set(section_config_name, entry.config_name, str(value))

                config.write(config_file)
                self.values_hash = self.get_values_hash()
        except Exception as e:
            # In case of error restore config from backup file
            tkinter.messagebox.showinfo("Exeption",
                                  "There was error during saving configuration. See console for more information.\n{}".format(
                                      str(e)))
            print((traceback.print_exc()))
            shutil.copy(config_filename + ".bak", config_filename)

        os.remove(config_filename + ".bak")

        if required_checking:
            tkinter.messagebox.showinfo("Saved", "Saving complete")

        return True

    def open_docs(self):
        webbrowser.open("http://www.aquaduct.pl/apidocs/valve/valve_config.html")

    def create_new_config_file_dialog(self):
        """
        Show dialog with create or choose existing file options.
        """
        window = tk.Toplevel(self.parent, padx=20)
        window.title("New file")

        directory_name = tk.StringVar()
        file_name = tk.StringVar()

        def select_dir():
            directory_name.set(askdirectory())
            window.lift()  # Move to foreground

        tkinter.ttk.Label(window, text="Directory:").grid(row=0, column=0)
        tkinter.ttk.Entry(window, textvariable=directory_name, state="readonly").grid(row=0, column=1)
        tkinter.ttk.Button(window, text="Choose directory", command=select_dir, style="File.TButton").grid(row=0, column=2)

        tkinter.ttk.Label(window, text="File name:").grid(row=1, column=0)
        tkinter.ttk.Entry(window, textvariable=file_name).grid(row=1, column=1)

        def create(dir_var, filename_var):
            path = dir_var.get() + os.path.sep + filename_var.get()
            with open(path, "w+"):
                self.config_filename.set(path)
                self.parent.title("{} - {}".format(self.title, path))

            window.destroy()

        create_callback = utils.CallbackWrapper(create, directory_name, file_name)
        tkinter.ttk.Button(window, text="Create", command=create_callback).grid(row=3, column=0, columnspan=3)

    def about(self):
        window = tk.Toplevel(self.parent)
        window.title("About")

        logo = tk.PhotoImage(file=get_img("logo.gif"))

        logo_label = tkinter.ttk.Label(window, image=logo)
        logo_label.image = logo
        logo_label.pack()

        content = """Aqua-Duct {}

ValveConfigurator

Copyright Tunneling Group \xa9 2018""".format(aquaduct.version_nice())

        tkinter.ttk.Label(window, text=content, justify=tk.CENTER).pack()
        tkinter.ttk.Label(window, text=aquaduct.__author__, justify=tk.CENTER).pack()

        hyperlink = tk.Label(window, text="http://www.tunnelinggroup.pl/", foreground="blue", justify=tk.CENTER)
        hyperlink.pack()
        hyperlink_cb = utils.CallbackWrapper(webbrowser.open_new, "http://www.tunnelinggroup.pl/")
        hyperlink.bind("<Button-1>", hyperlink_cb)

    def valve_run_dialog(self):
        """ Open dialog with options to run Valve """
        if self.config_filename.get() == "":
            tkinter.messagebox.showinfo("Config file", "You have to save or open existing config.")
            return
        elif self.values_hash != self.get_values_hash():
            result = tkinter.messagebox.askyesno("Unsaved config", "There are unsaved changed. Are you sure to continue?")

            if not result:
                return

        window = tk.Toplevel(self.parent, padx=20, pady=15)
        window.title("Run Valve")

        scrolled_frame = utils.VerticalScrolledFrame(window)
        scrolled_frame.pack(expand=1, fill=tk.BOTH)
        frame = scrolled_frame.interior
        frame.columnconfigure(0, weight=1)
        frame.columnconfigure(1, weight=1)

        command = tk.StringVar()
        args = {}

        def update_command(*a):
            cmd = "valve_run "
            for arg, var in args.items():
                if isinstance(var.get(), bool):
                    if var.get():
                        cmd += arg + " "
                elif var.get():
                    cmd += "{} {} ".format(arg, var.get())

            command.set(cmd)

        for row, entry in enumerate(defaults.VALVE_DEFAULTS.entries):
            var = utils.entry_factory(frame, row, entry.name, entry.default_values, entry.help_text)
            args[entry.config_name] = var
            var.input_var.trace("w", update_command)

        utils.Text(frame, textvariable=command, width=60, height=2).grid(row=999,
                                                                         column=0,
                                                                         columnspan=2,
                                                                         sticky="ew",
                                                                         pady=10)

        # Update field with opened config file
        args["-c"].set(self.config_filename.get())

        run_cb = utils.CallbackWrapper(os.system, command.get() + " & disown")
        run_button = tkinter.ttk.Button(frame, text="Run")
        run_button.bind("<Button-1>", run_cb)
        run_button.grid(row=1000, column=0, columnspan=2)

    def pond_run_dialog(self):
        """ Open dialog with options to run Valve """
        if self.config_filename.get() == "":
            tkinter.messagebox.showinfo("Config file", "You have to save or open existing config.")
            return
        elif self.values_hash != self.get_values_hash():
            result = tkinter.messagebox.askyesno("Unsaved config", "There are unsaved changed. Are you sure to continue?")

            if not result:
                return

        window = tk.Toplevel(self.parent, padx=20, pady=15)
        window.title("Run Pond")

        scrolled_frame = utils.VerticalScrolledFrame(window)
        scrolled_frame.pack(expand=1, fill=tk.BOTH)
        frame = scrolled_frame.interior
        frame.columnconfigure(0, weight=1)
        frame.columnconfigure(1, weight=1)

        command = tk.StringVar()
        args = {}

        def update_command(*a):
            cmd = "pond_run "
            for arg, var in args.items():
                if isinstance(var.get(), bool):
                    if var.get():
                        cmd += arg + " "
                elif var.get():
                    cmd += "{} {} ".format(arg, var.get())

            command.set(cmd)

        for row, entry in enumerate(defaults.POND_DEFAULTS.entries):
            var = utils.entry_factory(frame, row, entry.name, entry.default_values, entry.help_text)
            args[entry.config_name] = var
            var.input_var.trace("w", update_command)

        utils.Text(frame, textvariable=command, width=60, height=2).grid(row=999,
                                                                         column=0,
                                                                         columnspan=2,
                                                                         sticky="ew",
                                                                         pady=10)

        # Update field with opened config file
        args["-c"].set(self.config_filename.get())

        run_cb = utils.CallbackWrapper(os.system, command.get())
        run_button = tkinter.ttk.Button(frame, text="Run")
        run_button.bind("<Button-1>", run_cb)
        run_button.grid(row=1000, column=0, columnspan=2)

    def on_window_close(self):
        if self.get_values_hash() != self.values_hash:
            result = tkinter.messagebox.askyesno("Unsaved changes", "You have unsaved changes. Are you want to quit?")
            if result:
                self.parent.destroy()
        else:
            self.parent.destroy()


if __name__ == "__main__":
    root = tk.Tk()
    root.configure(background="white")
    root.resizable(1, 1)

    aq_icon = tk.PhotoImage(file=get_img("icon.gif"))
    root.tk.call('wm', 'iconphoto', root._w, aq_icon)

    s = tkinter.ttk.Style()

    s.theme_use("clam")

    ###
    # Global styles
    ###
    font_name = "TkDefaultFont"
    font_size = 8
    s.configure("TLabel", padding=5, background="#dcdad5", font=(font_name, font_size))
    s.configure("TButton", padding=5, background="#dcdad5", font=(font_name, font_size))
    s.configure("TMenubutton", padding=5, background="#dcdad5", font=(font_name, font_size))

    # Top label in every self.notebook tab
    s.configure("Title.TLabel", font=("TkDefaultFont", 16))

    ###
    # self.notebook style
    ###
    s.configure("TNotebook.Tab", width=20, anchor=tk.CENTER, font=("TkDefaultFont", 8))

    # s.configure("Tself.notebook", tabposition="wn")
    # s.map("Tself.notebook.Tab",
    # width=[("selected", 20)],
    # padding=[("selected", [0, 7])]
    # )

    ###
    # For recursive clustering frames
    ###
    s.configure("I.TFrame", background="#E6E4DF", relief=tk.SUNKEN)
    s.configure("I.TFrame.Label", background="#E6E4DF")
    s.configure("ITitle.TLabel", background="#E6E4DF", font=("TkDefaultFont", 14))

    ###
    # For hiding frames
    ###
    s.configure("HF.TFrame", background="#F0EEE9", relief=tk.RAISED)
    s.configure("HF.TFrame.Label", background="#F0EEE9", font=("TkDefaultFont", 9, "bold"))

    ###
    # Other
    ###
    s.configure("File.TButton", padding=0, font=("TkDefaultFont", 8))  # Loading file button
    s.configure("Configured.TLabel", padding=0, font=("TkDefaultFont", 12))  # Loading file button
    s.configure("R.TButton", background="red3", bordercolor="red3")  # Reset button

    ###
    # Highlighted
    ###
    s.configure("Highlighted.TFrame", background="red")

    app = ValveConfigApp(root)
    root.mainloop()



# -*- coding: utf-8 -*-

# Aqua-Duct, a tool facilitating analysis of the flow of solvent molecules in molecular dynamic simulations
# Copyright (C) 2018-2019  Tomasz Magdziarz <info@aquaduct.pl>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from aquaduct.traj.paths import GenericPaths
from collections import defaultdict
import operator
import numpy as np

from multiprocessing import Manager

from scipy import spatial


def get_spc(sp, window=None):
    '''
    :param sp: Single path like object or Generic path.
    :param tuple window: Optional frames window.
    :rtype: numpy.ndarray
    :return: Coordinates of path; to be used in pocket calculation.
    '''
    if isinstance(sp, GenericPaths):
        if window is None:
            return sp.coords
        f = np.array(sp.frames)
        i = (f >= window[0]) & (f <= window[1])
        return get_spc(sp)[i]
    else:
        if window is None:
            return sp.coords_cont
        f = np.array(sp.paths_cont)
        i = (f >= window[0]) & (f <= window[1])
        return get_spc(sp)[i]


def find_minmax(spaths, pbar=None):
    '''
    :param list spaths: List of single like path objects.
    :param pbar: Optional progress object providing next() method.
    :rtype: 2 element tuple of numpy.ndarray each of shape (3,)
    :return: Minimal and maximal boundaries of coordinates used in pocket calulations of spaths.
    '''
    minc = np.array([float('inf')] * 3)
    maxc = np.array([float('-inf')] * 3)
    for sp in spaths:
        sp_minc = get_spc(sp).min(0)
        minc_i = minc > sp_minc
        minc[minc_i] = sp_minc[minc_i]
        sp_maxc = get_spc(sp).max(0)
        maxc_i = maxc < sp_maxc
        maxc[maxc_i] = sp_maxc[maxc_i]
        if pbar:
            pbar.next()
    minc = np.floor(minc)
    maxc = np.ceil(maxc)
    return minc, maxc


def find_minmax_single(sp):
    coords = get_spc(sp)
    return coords.min(0), coords.max(0)


def find_minmax_map(spaths, pbar=None, map_fun=None):
    '''
    :param list spaths: List of single like path objects.
    :param pbar: Optional progress object providing next() method.
    :rtype: 2 element tuple of numpy.ndarray each of shape (3,)
    :return: Minimal and maximal boundaries of coordinates used in pocket calulations of spaths.
    '''
    minc = np.array([float('inf')] * 3)
    maxc = np.array([float('-inf')] * 3)
    if map_fun is None:
        map_fun = imap
    for sp_minc, sp_maxc in map_fun(find_minmax_single, spaths):
        minc_i = minc > sp_minc
        minc[minc_i] = sp_minc[minc_i]
        maxc_i = maxc < sp_maxc
        maxc[maxc_i] = sp_maxc[maxc_i]
        if pbar:
            pbar.next()
    minc = np.floor(minc)
    maxc = np.ceil(maxc)
    return minc, maxc


def find_edges(spaths, grid_size=1., pbar=None, map_fun=None):
    '''
    :param list spaths: List of single like path objects.
    :param float grid_size: Size of grid cell in A.
    :param pbar: Optional progress object providing next() method.
    :rtype: list of numpy.ndarrays
    :return: Edges of bins of grid spanning all submited paths.
    '''
    return [np.linspace(mi, ma, int((ma - mi) / grid_size) + 1) for mi, ma in
            zip(*find_minmax_map(spaths, pbar=pbar, map_fun=map_fun))]


class distribution_worker(object):
    def __init__(self, edges=None, window=None):
        self.edges = edges
        self.window = window

    def __call__(self, sp):
        return np.histogramdd(get_spc(sp, window=self.window), bins=self.edges)[0]


def distribution(spaths, grid_size=1., edges=None, window=None, pbar=None, map_fun=None):
    '''
    :param list spaths: List of single like path objects.
    :param float grid_size: Size of grid cell in A.
    :param list of numpy.ndarrays edges: Edges of bins of grid spanning all submited paths.
    :param tuple window: Optional frames window.
    :param pbar: Optional progress object providing next() method.
    :rtype tuple of numpy.ndarrays
    :return: Coordinates of pocket and number of points.
    '''
    maxc = np.array(list(map(max, edges)))
    minc = np.array(list(map(min, edges)))
    # H = np.zeros(map(int,map(np.ceil,(maxc - minc) / grid_size)))
    H = np.zeros(tuple([len(e) - 1 if len(e) > 1 else 1 for e in edges]))
    if map_fun is None:
        map_fun = map
    map_worker = distribution_worker(edges=edges, window=window)
    for h in map_fun(map_worker, spaths):
        H += h
        if pbar:
            pbar.next()
    mg = [ee[:-1] + (grid_size / 2.) for ee in edges]
    x, y, z = np.meshgrid(*mg, indexing='ij')
    pocket = H > 0
    return np.vstack((x[pocket], y[pocket], z[pocket])).T, H[pocket]


def outer_inner(H, threshold=None):
    '''
    :param numpy.ndarray H: Pocket distribution.
    :param float threshold: Percent value of max density which will be used to partition pocket into inner and outer.
    :return: Indices of outer and inner pocket.
    :rtype: tuple of numpy.ndarray
    '''
    if threshold:
        assert threshold <= 1.0, "Threshold cannot be higher that 1.0."
        assert threshold > 0.0, "Threshold cannot be equal or less than 0."
    if (H > 0).any():
        OI = H / H[H > 0].mean() if not threshold else H / (H[H > 0].max() * threshold)
        return OI < 1, OI >= 1
    return np.ones(H.shape) > 1, np.ones(H.shape) == 0  # fall back to outer pocket only


def windows(frames, windows=None, size=None):
    yield (0., frames - 1.)  # full window
    if windows:
        if size:
            begs = np.linspace(0, frames - size, windows)
            ends = np.array([b + size - 1 for b in begs])
            if ends[-1] > frames - 1:
                ends[-1] = frames - 1
        else:
            begs = np.linspace(0, frames, windows + 1)[:-1]
            ends = np.linspace(-1, frames - 1, windows + 1)[1:]
            if ends[-1] < frames - 1:
                ends[-1] = frames - 1
        for b, e in zip(begs, ends):
            yield np.floor(b), np.floor(e)


def sphere_radii(spaths, centers=None, radii=None, window=None, pbar=None, map_fun=None):
    H = np.zeros(len(centers), dtype=np.int32)

    if map_fun is None:
        map_fun = imap

    for sp in spaths:
        coords = get_spc(sp, window=window)
        D = spatial.distance.cdist(coords, centers)
        for nr, radius in enumerate(radii):
            H[nr] += np.count_nonzero(D[:, nr] <= radius)
        if pbar:
            pbar.next()
    return H


class sphere_radius_worker(object):
    def __init__(self, window, centers, radius):
        self.window = window
        self.centers = centers
        self.radius = radius

    def __call__(self, sp):
        coords = get_spc(sp, window=self.window)
        D = spatial.distance.cdist(coords, self.centers) <= self.radius
        return np.count_nonzero(D, 0)


class sphere_radius_worker_lowmem(object):
    def __init__(self, window, centers, radius):
        self.window = window
        self.centers = centers
        self.radius = radius

    def __call__(self, sp):
        coords = get_spc(sp, window=self.window)
        g = (int(np.count_nonzero(spatial.distance.cdist(coords, [c]) <= self.radius)) for c in self.centers)
        return np.fromiter(g, dtype=np.int32)


def sphere_radius(spaths, centers=None, radius=2., window=None, pbar=None, map_fun=None):
    H = np.zeros(len(centers), dtype=np.int32)

    if map_fun is None:
        map_fun = imap
    map_worker = sphere_radius_worker_lowmem(window, centers, radius)
    for h in map_fun(map_worker, spaths):
        H += h
        if pbar:
            pbar.next()

    return H


class sphere_density_raw_worker(object):
    def __init__(self, pbar_queue, mol_name, radius, centers, frames):
        """
        Worker used to calculate energy per each trajectory.

        :param pbar_queue: Progress bar instance
        :type pbar_queue: class:`aquaduct.clui.SimpleProgressBar`
        :param mol_name: Name of molecule which amount will be calculated around points.
        :type mol_name: str
        :param radius: Radius of selection for each point.
        :type radius: float
        :param centers: Points of path
        :type centers: Iterator
        :param frames: Collection of frame numbers
        :type frames: Iterator
        """
        self.pbar_queue = pbar_queue
        self.mol_name = mol_name
        self.radius = radius
        self.centers = centers
        self.frames = frames

    def __call__(self, traj_reader):
        """
        Worker

        :param traj_reader: Trajectory reader
        :type traj_reader: Iterator
        :return: Array with density for each point
        :rtype: numpy.ndarray
        """
        H = np.zeros(len(self.centers))

        traj_reader = traj_reader.open()

        # Iterating over trajectory reader starts from 0.
        # It's needed to add value of first frame to check if frame is within frames range
        base = traj_reader.window.start
        for frame in traj_reader.iterate():

            # Skip if frame is not within range
            if frame + base not in self.frames:
                continue

            # Calculate density of points for frame
            for i, c in enumerate(self.centers):
                selection_str = "({}) and (point {} {} {} {})".format(self.mol_name, *(tuple(c) + (self.radius,)))
                traj_reader.parse_selection(selection_str)
                H[i] += len(list(traj_reader.parse_selection(selection_str).residues().ids()))

            # Indicate update of progress bar
            self.pbar_queue.put(1)

        return H


def sphere_density_raw(trajs, mol_name, centers, radius, pool, window=None, pbar=None):
    """
    Calculate density of sphere with specified radius and with center in each point.

    :param trajs: Collection of trajectory readers
    :type trajs: class:`aquaduct.sandwich.ReaderTraj`
    :param mol_name: Name of molecule which amount will be calculated around points.
    :type mol_name: str
    :param radius: Radius of selection for each point.
    :type radius: float
    :param centers: Points of path
    :type centers: Iterator
    :param pool: Preconfigured Pool instance
    :type pool: Preconfigured Pool instance
    :return: Density for each center.
    :rtype: numpy.ndarray
    """

    # Setup for queue and worker
    pbar_queue = Manager().Queue()
    worker = sphere_density_raw_worker(pbar_queue, mol_name, radius, centers, list(range(int(window[0]), int(window[1]))))

    # Container for collected data from workers
    D = []

    r = pool.map_async(worker, trajs, callback=D.append)

    # Incrementing progress bar using queued values
    for p in iter(pbar_queue.get, None):
        pbar.next(step=p)

        if pbar.current == pbar.maxval:
            break

    r.wait()

    # Return summed results from processes
    return np.array(D[0]).sum(axis=0)


def hot_spots(H):
    return hot_spots_his(H)


def hot_spots_his(H, bins=(5, 101)):
    bn = []
    for b in range(*bins):
        his = np.histogram(H, bins=b)
        if 0 in his[0]:
            i = int(np.argwhere(his[0] == 0)[0])
            bn.append((b, sum(his[0][i:])))
    if len(bn):
        i = np.argmax(np.array(bn)[:, 1])
        b = bn[i][0]
        his = np.histogram(H, bins=b)
        hs = np.argwhere(his[0] == 0)
        return his[1][hs][0]



 # -*- coding: utf-8 -*-

# Aqua-Duct, a tool facilitating analysis of the flow of solvent molecules in molecular dynamic simulations
# Copyright (C) 2018-2019  Tomasz Magdziarz <info@aquaduct.pl>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

AVAILABLE_SMOOTHING_METHODS = ['window', 'dwin', 'awin', 'mss', 'window_mss', 'dwin_mss', 'awin_mss', 'savgol']
